"""Software for fitting HMP on EEG/MEG data."""


from . import models
from . import utils
from . import visu
from . import resample
from . import clusters
from . import loocv

__all__ = ["models", "simulations", "utils","visu", "clusters","mcca","loocv"]


'''

'''

import numpy as np
from scipy.special import gamma as gamma_func
from scipy.stats import lognorm
import xarray as xr
import multiprocessing as mp
import itertools
import pandas as pd
from pandas import MultiIndex
import warnings
from warnings import warn, filterwarnings
from seaborn.algorithms import bootstrap
from hmp import mcca
import json
import mne
import os

filterwarnings('ignore', 'Degrees of freedom <= 0 for slice.', )#weird warning, likely due to nan in xarray, not important but better fix it later 
filterwarnings('ignore', 'Mean of empty slice')#When trying to center all-nans trials

def gamma_scale_to_mean(scale, shape): 
    return scale*shape
def gamma_mean_to_scale(mean, shape): 
    return mean/shape

def logn_scale_to_mean(scale, shape): 
    return np.exp(scale+shape**2/2)
def logn_mean_to_scale(mean, shape):
    return np.exp(np.log(mean)-(shape**2/2))

def wald_scale_to_mean(scale, shape): 
    return scale*shape
def wald_mean_to_scale(mean, shape): 
    return mean/shape

def weibull_scale_to_mean(scale, shape):
    return scale*gamma_func(1+1/shape)
def weibull_mean_to_scale(mean, shape): 
    return mean/gamma_func(1+1/shape)

def maxb_scale_to_mean(scale, shape):
    return 2*scale*np.sqrt(2/np.pi)
def maxb_mean_to_scale(mean, shape): 
    return mean/2/np.sqrt(2/np.pi)

def ray_scale_to_mean(scale, shape):
    return scale*np.sqrt(np.pi/2)
def ray_mean_to_scale(mean, shape): 
    return mean/np.sqrt(np.pi/2)

def halfn_scale_to_mean(scale, shape):
    return (scale*np.sqrt(2))/np.sqrt(np.pi)
def halfn_mean_to_scale(mean, shape): 
    return mean/np.sqrt(2/np.pi)

def fisk_scale_to_mean(scale, shape):
    return  (scale*np.pi/shape)/np.sin(np.pi/shape)
def fisk_mean_to_scale(mean, shape): 
    return  shape*(mean*np.sin(np.pi/shape))/np.pi

def uniform_scale_to_mean(scale, shape): 
    return (scale-shape)/2
def uniform_mean_to_scale(mean, shape): 
    return 2*(mean-shape)

def read_mne_data(pfiles, event_id=None, resp_id=None, epoched=False, sfreq=None, 
                 subj_idx=None, metadata=None, events_provided=None, rt_col='rt', rts=None,
                 verbose=True, tmin=-.2, tmax=5, offset_after_resp = 0, 
                 high_pass=None, low_pass = None, pick_channels = 'eeg', baseline=(None, 0),
                 upper_limit_RT=np.inf, lower_limit_RT=0, reject_threshold=None, scale=1, reference=None, ignore_rt=False):
    ''' 
    Reads EEG/MEG data format (.fif or .bdf) using MNE's integrated function .
    
    Notes: 
    - Only EEG or MEG data are selected (other channel types are discarded)
    - All times are expressed on the second scale.
    - If multiple files in pfiles the data of the group is read and seqentially processed.
    - For non epoched data: Reaction Times are only computed if response trigger is in the epoch window (determined by tmin and tmax)
    
    Procedure:
    if data not already epoched:
        0.1) the data is filtered with filters specified in low_pass and high_pass. Parameters of the filter are
        determined by MNE's filter function.
        0.2) if no events is provided, detect events in stimulus channel and keep events with id in event_id and resp_id.
        0.3) eventual downsampling is performed if sfreq is lower than the data's sampling frequency. The event structure is
        passed at the resample() function of MNE to ensure that events are approriately timed after downsampling.
        0.4) epochs are created based on stimulus onsets (event_id) and tmin and tmax. Epoching removes any epoch where a 
        'BAD' annotiation is present and all epochs where an channel exceeds reject_threshold. Epochs are baseline 
        corrected from tmin to stimulus onset (time 0).)
    1) Reaction times (RT) are computed based on the sample difference between onset of stimulus and response triggers. 
        If no response event happens after a stimulus or if RT > upper_limit_RT & < upper_limit_RT, RT is 0.
    2) all the non-rejected epochs with positive RTs are cropped to stimulus onset to stimulus_onset + RT.
    
    Parameters
    ----------
    pfiles : str or list
        list of EEG files to read,
    event_id : dict
        Dictionary containing the correspondance of named condition [keys] and event code [values]
    resp_id : dict
        Dictionary containing the correspondance of named response [keys] and event code [values]
    sfreq : float
        Desired sampling frequency
    to_merge_id: dict
        Dictionary containing the correspondance of named condition [keys] and event code [values] that needs to be
        merged with the stimuli event in event_id
    subj_idx : list
        List of subject names
    events_provided : float
        np.array with 3 columns -> [samples of the event, initial value of the channel, event code]. To use if the
        automated event detection method of MNE is not appropriate 
    verbose : bool
        Whether to display MNE's message
    tmin : float
        Time taken before stimulus onset to compute baseline
    tmax : float
        Time taken after stimulus onset
    offset_after_resp : float
        Time taken after onset of the response in seconds
    low_pass : float
        Value of the low pass filter
    high_pass : float
        Value of the high pass filter
    pick_channels: list 
        'eeg' (default) to keep only EEG channels or  list of channel names to keep
    baseline : tuple
        Time values to compute the baseline and substract to epoch data (usually some time before stimulus onset)
    upper_limit_RT : float
        Upper limit for RTs. Longer RTs are discarded
    lower_limit_RT : float
        Lower limit for RTs. Shorter RTs are discarded
    reject_threshold : float
        Rejection threshold to apply when creating epochs, expressed in microvolt
    scale: float
        Scale to apply to the RT data (e.g. 1000 if ms)
    reference:
        What reference to use (see MNE documentation), if None, keep the existing one
    ignore_rt: bool
        Use RT to parse the epochs (False, Default) or ignore the RT and parse up to tmaxx in epochs (True)
    Returns
    -------
    epoch_data : xarray
        Returns an xarray Dataset with all the data, events, channels, participant. 
        All eventual participant/channels naming and epochs index are kept. 
        The choosen sampling frequnecy is stored as attribute.
    '''
    import mne
    dict_datatype = {False:'continuous', True:'epoched'}
    epoch_data = [] 
    if isinstance(pfiles,str):#only one participant
        pfiles = [pfiles]
    if not subj_idx:
        subj_idx = ["S"+str(x) for x in np.arange(len(pfiles))]
    if isinstance(subj_idx,str):
        subj_idx = [subj_idx]
    if upper_limit_RT<0 or lower_limit_RT<0:
        raise ValueError('Limit to RTs cannot be negative')
    y = 0
    if metadata is not None:
        if len(pfiles)> 1 and len(metadata) != len(pfiles):
            raise ValueError(f'Incompatible dimension between the provided metadata {len(metadata)} and the number of eeg files provided {len(pfiles)}')
    else:
        metadata_i = None
    ev_i = 0 #syncing up indexing between event and raw files
    for participant in pfiles:

        print(f"Processing participant {participant}'s {dict_datatype[epoched]} {pick_channels}")

        # loading data
        if epoched == False:# performs epoching on raw data
            if '.fif' in participant:
                data = mne.io.read_raw_fif(participant, preload=False, verbose=verbose)
            elif '.bdf' in participant:
                data = mne.io.read_raw_bdf(participant, preload=False, verbose=verbose)
            else:
                raise ValueError(f'Unknown EEG file format for participant {participant}')
            if sfreq is None: 
                sfreq = data.info['sfreq']

            if 'response' not in list(resp_id.keys())[0]:
                resp_id = {f'response/{k}': v for k, v in resp_id.items()}
            if events_provided is None:
                try:
                    events = mne.find_events(data, verbose=verbose, min_duration = 1 / data.info['sfreq'])
                except:
                    events = mne.events_from_annotations(data, verbose=verbose)[0]
                if events[0,1] > 0:#bug from some stim channel, should be 0 otherwise indicates offset in the trggers
                    print(f'Correcting event values as trigger channel has offset {np.unique(events[:,1])}')
                    events[:,2] = events[:,2]-events[:,1]#correction on event value             
                events_values = np.concatenate([np.array([x for x in event_id.values()]), np.array([x for x in resp_id.values()])])
                events = np.array([list(x) for x in events if x[2] in events_values])#only keeps events with stim or response
                events_stim = np.array([list(x) for x in events if x[2] in event_id.values()])#only stim
            else:
                if len(events_provided[0]) == 3:
                    events_provided = events_provided[np.newaxis]
                    events = events_provided[y]
                else:#assumes stacked event files
                    events = events_provided[ev_i]
                    ev_i += 1
            if reference is not None:
                data = data.set_eeg_reference(reference)
            data = _pick_channels(pick_channels,data, stim=True)
            data.load_data()

            if sfreq < data.info['sfreq']:#Downsampling
                print(f'Downsampling to {sfreq} Hz')
                decim = np.round( data.info['sfreq'] / sfreq).astype(int)
                obtained_sfreq = data.info['sfreq'] / decim
                low_pass = obtained_sfreq / 3.1
            else: 
                decim = 1
                if sfreq > data.info['sfreq']+1:
                    warn(f'Requested higher frequency {sfreq} than found in the EEG data, no resampling is performed')
            if high_pass is not None or low_pass is not None:
                data.filter(high_pass, low_pass, fir_design='firwin', verbose=verbose)
            combined =  {**event_id, **resp_id}#event_id | resp_id 
            stim = list(event_id.keys())
            
            if verbose:
                print(f'Creating epochs based on following event ID :{np.unique(events[:,2])}')

            offset_after_resp_samples = int(offset_after_resp*sfreq)
            if metadata is None:
                metadata_i, meta_events, event_id = mne.epochs.make_metadata(
                    events=events, event_id=combined, tmin=tmin, tmax=tmax,
                    sfreq=data.info["sfreq"], row_events=stim, keep_first=["response"])
                metadata_i = metadata_i[["event_name","response"]]#only keep event_names and rts
            else:
                metadata_i = metadata[y]
            epochs = mne.Epochs(data, meta_events, event_id, tmin, tmax, proj=False,
                    baseline=baseline, preload=True, picks=pick_channels, decim=decim,
                    verbose=verbose, detrend=None, on_missing = 'warn', event_repeated='drop',
                    metadata=metadata_i, reject_by_annotation=True, reject=reject_threshold)
            epochs.metadata.rename({'response':'rt'}, axis=1, inplace=True)
            metadata_i = epochs.metadata
        else:
            if '.fif' in participant:
                epochs = mne.read_epochs(participant, preload=True, verbose=verbose)
                if high_pass is not None or low_pass is not None:
                    epochs.filter(high_pass, low_pass, fir_design='firwin', verbose=verbose)
                if sfreq is None: 
                    sfreq = epochs.info['sfreq']
                elif sfreq  < epochs.info['sfreq']:
                    if verbose:
                        print(f'Resampling data at {sfreq}')
                    epochs = epochs.resample(sfreq)
            else:
                raise ValueError('Incorrect file format')
            if reference is not None:
                epochs = epochs.set_eeg_reference(reference)
            _pick_channels(pick_channels,epochs, stim=False)
            if metadata is None:
                try:
                    metadata_i = epochs.metadata#accounts for dropped epochs
                except:
                    raise ValueError('Missing metadata in the epoched data')
            elif isinstance(metadata, pd.DataFrame):
                if len(pfiles)>1:
                    metadata_i = metadata[y].copy()#TODO, better account for participant's wide provided metadata
                else:
                    metadata_i = metadata.copy()
            else:
                raise ValueError('Metadata should be a pandas data-frame as generated by mne or be contained in the passed epoch data')
        if upper_limit_RT == np.inf:
            upper_limit_RT = epochs.tmax+1*(1/sfreq)
        if ignore_rt:
            metadata_i[rt_col] = epochs.tmax
        offset_after_resp_samples = np.rint(offset_after_resp*sfreq).astype(int)
        valid_epoch_index = [x for x,y in enumerate(epochs.drop_log) if len(y) == 0]
        try:#Differences among MNE's versions
            data_epoch = epochs.get_data(copy=False)#preserves index
        except:
            data_epoch = epochs.get_data()#preserves index
        rts = metadata_i[rt_col]
        if isinstance(metadata_i, pd.DataFrame):
            if len(metadata_i) > len(data_epoch):#assumes metadata contains rejected epochs
                metadata_i = metadata_i.loc[valid_epoch_index]
                rts = metadata_i[rt_col]
            try:
                rts = rts/scale
            except:
                raise ValueError(f'Expected column named {rt_col} in the provided metadata file, alternative names can be passed through the rt_col parameter')
        elif rts is None:
            raise ValueError(f'Expected either a metadata Dataframe or an array of Reaction Times')
        rts_arr = np.array(rts)
        if verbose:
            print(f'Applying reaction time trim to keep RTs between {lower_limit_RT} and {upper_limit_RT} seconds')
        rts_arr[rts_arr > upper_limit_RT] = 0 #removes RT above x sec
        rts_arr[rts_arr < lower_limit_RT] = 0 #removes RT below x sec, important as determines max events
        rts_arr[np.isnan(rts_arr)] = 0#too long trial
        rts_arr = np.rint(rts_arr*sfreq).astype(int)
        if verbose:
            print(f'{len(rts_arr[rts_arr > 0])} RTs kept of {len(rts_arr)} clean epochs')
        triggers = metadata_i.iloc[:,0].values#assumes first col is trigger
        cropped_data_epoch = np.empty([len(rts_arr[rts_arr>0]), len(epochs.ch_names), max(rts_arr)+offset_after_resp_samples])
        cropped_data_epoch[:] = np.nan
        cropped_trigger = []
        epochs_idx = []
        j = 0
        time0 = epochs.time_as_index(0)[0]
        for i in range(len(data_epoch)):
            if rts_arr[i] > 0:
                cropped_trigger.append(triggers[i])
            #Crops the epochs to time 0 (stim onset) up to RT
                cropped_data_epoch[j,:,:rts_arr[i]+offset_after_resp_samples] = \
                (data_epoch[i,:,time0:time0+rts_arr[i]+offset_after_resp_samples])
                epochs_idx.append(valid_epoch_index[i])#Keeps trial number
                j += 1
        x = 0
        while np.isnan(cropped_data_epoch[-1]).all():#Weird bug I guess it is perhps due to too long last epoch? update: cannot reproduce
            cropped_data_epoch = cropped_data_epoch[:-1]
            x += 1
        if x > 0:
            print(f'RTs > 0 longer than expected ({x})')
        print(f'{len(cropped_data_epoch)} trials were retained for participant {participant}')
        if verbose:
            print(f'End sampling frequency is {sfreq} Hz')

        epoch_data.append(hmp_data_format(cropped_data_epoch, epochs.info['sfreq'], None, offset_after_resp_samples, epochs=[int(x) for x in epochs_idx], channels = epochs.ch_names, metadata = metadata_i))

        y += 1
    epoch_data = xr.concat(epoch_data, dim = xr.DataArray(subj_idx, dims='participant'),
                          fill_value={'event':'', 'data':np.nan})
    epoch_data = epoch_data.assign_attrs(lowpass=epochs.info['lowpass'], highpass=epochs.info['highpass'],
                                         lower_limit_RT=lower_limit_RT,  upper_limit_RT=upper_limit_RT, 

)
    return epoch_data

def _pick_channels(pick_channels,data,stim=True):
    if isinstance(pick_channels, list):
        try:
            data = data.pick(pick_channels)
        except:
            raise ValueError('incorrect channel pick specified')
    elif pick_channels == 'eeg' or pick_channels == 'meg':
            data = data.pick(pick_channels)
    else:
         raise ValueError('incorrect channel pick specified')
    return data

def parsing_epoched_eeg(data, rts, conditions, sfreq, start_time=0, offset_after_resp=0.1):
    '''
    Function to parse epochs and crop them to start_time (usually stimulus onset so 0) up to the reaction time of the trial.
    The returned object is a xarray Dataset allowing further processing using built-in methods

    Importantly 
    1) if you are considering some lower or upper limit on the RTs you should replace values outside of these ranges
    by np.nan (e.g. rts[rts < 200] = np.nan) or 0
    2) RTs and conditions need to be ordered in the same way as the epochs
    
    
    Parameters
    ----------
    data: pandas.dataframe
        pandas dataframe with columns time (in milliseconds), epoch number and one column for each channel 
        (column name will be taken as channel names)
    rts: list or 1d array
        list of reaction times in milliseconds for each epoch 
    epoch_index: list or 1d array
        number of the index (important for eventual dropped trials during the prepro154,cessing
    channel_index: list or 1d array
        list of name of the channels
    condition: list or 1d array
        list of condition associated with each epoch
    sfreq: float
        sampling frequency of the data
    start_time: float
        time defining the onset of a trial (default is 0 as trial usually defined from event/stimulus onset) in milliseconds
    offset_after_resp: float
        eventual time added after the response (e.g. if we expect later components) in milliseconds
    '''
    tstep = 1000/sfreq#time step
    offset_after_resp_samples = int(offset_after_resp/tstep)
    epoch_size = len(data.time.unique())*tstep
    if not isinstance(rts, np.ndarray):
        try:#pandas or xarray
            rts = rts.values
        except:
            raise ValueError('RTs should either be a numpy array or a pandas serie')
    if not isinstance(conditions, np.ndarray):
        try:#pandas or xarray
            conditions = conditions.values
        except:
            raise ValueError('Conditions should either be a numpy array or a pandas serie')
    if any(t > epoch_size for t in rts):
        print('Zeroing out RTs longer than epoch size, you might want to zero out rts that are outside of the range of interest')
        rts[rts > epoch_size] = 0
    rts[np.isnan(rts)] = 0
    rts = np.array([int(x) for x in rts/tstep])
    data = data[data.time >= start_time]#remove all baseline values
    epochs = data.epoch.unique()
    rts = rts[epochs]
    conditions = conditions[epochs]
    times = data.time.unique()
    data = data.drop(columns=['time', 'epoch'])
    #channel names/columns are assumed to be remaining columns affter removing time and epoch columns
    channel_columns = [x for x in data.columns]
    data = data.values.flatten()
    data = data.reshape((len(epochs), len(times),len(channel_columns)))
    data = np.swapaxes(data,1,2)

    nan_con = np.array([i for i,x in enumerate(conditions) if isinstance(x, float) and np.isnan(x)])
    if len(nan_con) > 0:
        print(f'NaN present in condition array, removing associated epoch and RT ({nan_con})')
        data = np.delete(data, nan_con, axis=0)
        conditions = np.delete(conditions, nan_con, axis=0)
        rts = np.delete(rts, nan_con, axis=0)
        epochs =  np.delete(epochs, nan_con, axis=0)
    epoch = 0
    for rt in rts:
        if rt == 0:
            data[epoch,:,:] = np.nan
            conditions[epoch] = ''
            #rts[epoch] = None#np.nan#np.delete(rts, epoch, axis=0)
            #epochs[epoch] = np.nan#np.delete(epochs, epoch, axis=0)
        epoch += 1
    cropped_conditions = []
    epoch_idx = []
    cropped_data_epoch = np.empty([len(epochs), len(channel_columns), max(rts)+offset_after_resp_samples])
    cropped_data_epoch[:] = np.nan
    j = 0
    for epoch in np.arange(len(data)):
        #Crops the epochs up to RT
        cropped_data_epoch[j,:,:rts[epoch]+offset_after_resp_samples] = \
        (data[epoch,:,:rts[epoch]+offset_after_resp_samples])
        j += 1
    print(f'Totaling {len(cropped_data_epoch)} valid trials')

    data_xr = hmp_data_format(cropped_data_epoch, sfreq, conditions, offset_after_resp_samples, epochs=epochs, channels = channel_columns)
    return data_xr

def hmp_data_format(data, sfreq, events=None, offset=0, participants=[], epochs=None, channels=None, metadata=None):

    '''
    Converting 3D matrix with dimensions (participant) * trials * channels * sample into xarray Dataset
    
    Parameters
    ----------
    data : ndarray
        4/3D matrix with dimensions (participant) * trials * channels * sample  
    events : ndarray
        np.array with 3 columns -> [samples of the event, initial value of the channel, event code]. To use if the
        automated event detection method of MNE is not appropriate 
    sfreq : float
        Sampling frequency of the data
    participants : list
        List of participant index
    epochs : list
        List of epochs index
    channels : list
        List of channel index
    '''
    if len(np.shape(data)) == 4:#means group
        n_subj, n_epochs, n_channels, n_samples = np.shape(data)
    elif len(np.shape(data)) == 3:
        n_epochs, n_channels, n_samples = np.shape(data)
        n_subj = 1
    else:
        raise ValueError(f'Unknown data format with dimensions {np.shape(data)}')
    if channels is None:
        channels = np.arange(n_channels)
    if epochs is None:
         epochs = np.arange(n_epochs)
    if n_subj < 2:
        data = xr.Dataset(
                {
                    "data": (["epochs", "channels", "samples"],data),
                },
                coords={
                    "epochs" :epochs,
                    "channels":  channels,
                    "samples": np.arange(n_samples)
                },
                attrs={'sfreq':sfreq,'offset':offset}
                )
    else:
        data = xr.Dataset(
                {
                    "data": (['participant',"epochs", "channels", "samples"],data),
                },
                coords={
                    'participant':participants,
                    "epochs" :epochs,
                    "channels":  channels,
                    "samples": np.arange(n_samples)
                },
                attrs={'sfreq':sfreq,'offset':offset}
                )
    if metadata is not None:
        metadata = metadata.loc[epochs]
        metadata = metadata.to_xarray()
        metadata = metadata.rename_dims({'index':'epochs'})
        metadata = metadata.rename_vars({'index':'epochs'})
        data = data.merge(metadata)
        data = data.set_coords(list(metadata.data_vars))
    if events is not None:
        data['events'] = xr.DataArray(
            events,
            dims=("participant", "epochs"),
            coords={"participant": participants, "epochs": epochs})
        data = data.set_coords('events')
    return data

def _standardize(x):
    '''
    Scaling variances to mean variance of the group
    '''
    return ((x.data / x.data.std(dim=...))*x.mean_std)

def _center(data):
    '''
    center the data
    '''
    mean_last_dim = np.nanmean(data.values, axis=-1)
    mean_last_dim_expanded = np.expand_dims(mean_last_dim, axis=-1)
    centred = data.values - mean_last_dim_expanded
    data.values = centred

    return data

def zscore_xarray(data):
    '''
    zscore of the data in an xarray, avoiding any nans
    '''
    if isinstance(data, xr.Dataset):#if no PCA
        data = data.data
    non_nan_mask = ~np.isnan(data.values)
    if non_nan_mask.any(): #if not everything is nan, calc zscore
        data.values[non_nan_mask] = (data.values[non_nan_mask] - data.values[non_nan_mask].mean()) / data.values[non_nan_mask].std()
    return data

def compute_ci(times):
    '''
    Compute confidence intervals
    '''
    return np.abs(np.squeeze([np.nanpercentile(bootstrap(times), q=[2.5,97.5])]) - np.mean(times))


def stack_data(data, subjects_variable='participant', channel_variable='component', single=False):
    '''
    Stacks the data going from format [participant * epochs * samples * channels] to [samples * channels]
    with sample indexes starts and ends to delimitate the epochs.
    
    
    Parameters
    ----------
    data : xarray
        unstacked xarray data from transform_data() or anyother source yielding an xarray with dimensions 
        [participant * epochs * samples * channels] 
    subjects_variable : str
        name of the dimension for subjects ID
    single : bool 
        Whether participant is unique (True) or a group of participant (False)
    
    Returns
    -------
    data : xarray.Dataset
        xarray dataset [samples * channels]
    '''    
    if isinstance(data, (xr.DataArray,xr.Dataset)) and 'component' not in data.dims:
        data = data.rename_dims({'channels':'component'})
    if "participant" not in data.dims:
        data = data.expand_dims("participant")
    data = data.stack(all_samples=['participant','epochs',"samples"]).dropna(dim="all_samples")
    return data

def _filtering(data, filter, sfreq):
    print("NOTE: filtering at this step is suboptimal, filter before epoching if at all possible, see")
    print("also https://mne.tools/stable/auto_tutorials/preprocessing/30_filtering_resampling.html")
    from mne.filter import filter_data

    lfreq, hfreq = filter
    n_participant, n_epochs, _, _ = data.data.values.shape
    for pp in range(n_participant):
        for trial in range(n_epochs):

            dat = data.data.values[pp, trial, :, :]

            if not np.isnan(dat).all():
                dat = dat[:,~np.isnan(dat[0,:])] #remove nans

                #pad by reflecting the whole trial twice
                trial_len = dat.shape[1] * 2
                dat = np.pad(dat, ((0,0),(trial_len,trial_len)), mode='reflect')

                #filter
                dat = filter_data(dat, sfreq, lfreq, hfreq, verbose=False)

                #remove padding
                dat = dat[:,trial_len:-trial_len]
                data.data.values[pp, trial, :, :dat.shape[1]] = dat
        return data

def _pca(pca_ready_data, n_comp, channels):
    from sklearn.decomposition import PCA
    # pca_ready_data = pca_ready_data.transpose(...,'channels')
    if n_comp == None:
        import matplotlib.pyplot as plt
        n_comp = np.shape(pca_ready_data)[0]-1
        fig, ax = plt.subplots(1,2, figsize=(.2*n_comp, 4))
        pca = PCA(n_components=n_comp, svd_solver='full')#selecting Principale components (PC)
        pca.fit(pca_ready_data)

        ax[0].plot(np.arange(pca.n_components)+1, pca.explained_variance_ratio_,'.-')
        ax[0].set_ylabel('Normalized explained variance')
        ax[0].set_xlabel('Component')
        ax[1].plot(np.arange(pca.n_components)+1, np.cumsum(pca.explained_variance_ratio_),'.-')
        ax[1].set_ylabel('Cumulative normalized explained variance')
        ax[1].set_xlabel('Component')
        plt.tight_layout()
        plt.show()
        n_comp = int(input(f'How many PCs (95 and 99% explained variance at component n{np.where(np.cumsum(pca.explained_variance_ratio_) >= .95)[0][0]+1} and n{np.where(np.cumsum(pca.explained_variance_ratio_) >= .99)[0][0]+1}; components till n{np.where(pca.explained_variance_ratio_ >= .01)[0][-1]+1} explain at least 1%)?'))

    pca = PCA(n_components=n_comp, svd_solver='full')#selecting Principale components (PC)
    pca.fit(pca_ready_data)
    #Rebuilding pca PCs as xarray to ease computation
    coords = dict(channels=("channels", channels),
                 component=("component", np.arange(n_comp)))
    pca_weights = xr.DataArray(pca.components_.T, dims=("channels","component"), coords=coords)
    return pca_weights

def transform_data(epoch_data, participants_variable="participant", apply_standard=False, averaged=False, apply_zscore='trial', zscore_acrossPCs=False, method='pca', cov=True, centering=True, n_comp=None, n_ppcas=None, pca_weights=None, bandfilter=None, mcca_reg=0):
    '''
    Adapts EEG epoched data (in xarray format) to the expected data format for hmps. 
    First this code can apply standardization of individual variances (if apply_standard=True).
    Second, a spatial PCA on the average variance-covariance matrix is performed (if method='pca', more methods in development)
    Third,stacks the data going from format [participant * epochs * samples * channels] to [samples * channels]
    Last, performs z-scoring on each epoch and for each principal component (PC), or for each participant and PC,
    or across all data for each PC.
        
    Parameters
    ----------
    data : xarray
        unstacked xarray data from transform_data() or anyother source yielding an xarray with dimensions 
        [participant * epochs * samples * channels] 
    participants_variable : str
        name of the dimension for participants ID
    apply_standard : bool 
        Whether to apply standardization of variance between participants, recommended when they are few of them (e.g. < 10)
    averaged : bool
        Applying the pca on the averaged ERP (True) or single trial ERP (False, default). No effect if cov = True
    apply_zscore : str 
        Whether to apply z-scoring and on what data, either None, 'all', 'participant', 'trial', for zscoring across all data, by participant, or by trial, respectively. If set to true, evaluates to 'trial' for backward compatibility.
    method : str
        Method to apply, 'pca' or 'mcca'
    cov : bool
        Wether to apply the pca/mcca to the variance covariance (True, default) or the epoched data
    n_comp : int
        How many components to select from the PC space, if None plots the scree plot and a prompt requires user
        to specify how many PCs should be retained
    n_ppcas : int
        If method = 'mcca', controls the number of components retained for the by-participant PCAs
    pca_weigths : xarray
        Weights of a PCA to apply to the data (e.g. in the resample function)
    bandfilter: None | (lfreq, hfreq) 
        If none, no filtering is appliedn. If tuple, data is filtered between lfreq-hfreq.
        NOTE: filtering at this step is suboptimal, filter before epoching if at all possible, see
              also https://mne.tools/stable/auto_tutorials/preprocessing/30_filtering_resampling.html
    mcca_reg: float
        regularization used for the mcca computation (see mcca.py)

    Returns
    -------
    data : xarray.Dataset
        xarray dataset [n_samples * n_comp] data expressed in the PC space, ready for HMP fit
    '''
    data = epoch_data.copy(deep=True)
    if isinstance(data, xr.DataArray):
        raise ValueError('Expected a xarray Dataset with data and event as DataArrays, check the data format')
    if not apply_zscore in ['all', 'participant', 'trial'] and not isinstance(apply_zscore,bool):
        raise ValueError('apply_zscore should be either a boolean or one of [\'all\', \'participant\', \'trial\']')
    assert np.sum(np.isnan(data.groupby('participant', squeeze=False).mean(['epochs','samples']).data.values)) == 0,\
        'at least one participant has an empty channel'
    if method == 'mcca' and data.sizes['participant'] == 1:
        raise ValueError('MCCA cannot be applied to only one participant')
    sfreq = data.sfreq
    if bandfilter:
        data = _filtering(data, bandfilter, sfreq)
    if apply_standard:
        if 'participant' not in data.dims or len(data.participant) == 1:
            warn('Requested standardization of between participant variance yet no participant dimension is found in the data or only one participant is present. No standardization is done, set apply_standard to False to avoid this warning.')
        else:
            mean_std = data.groupby(participants_variable, squeeze=False).std(dim=...).data.mean()
            data = data.assign(mean_std=mean_std.data)
            data = data.groupby(participants_variable, squeeze=False).map(_standardize)
    if isinstance(data, xr.Dataset):#needs to be a dataset if apply_standard is used
            data = data.data
    if centering or method=='mcca':
        data = _center(data)
    if apply_zscore == True:
        apply_zscore = 'trial' #defaults to trial
    data = data.transpose('participant','epochs','channels','samples')
    if method == 'pca':
        if pca_weights is None:
            if cov:
                indiv_data = np.zeros((data.sizes['participant'], data.sizes['channels'], data.sizes['channels']))
                for i in range(data.sizes['participant']):
                    x_i = np.squeeze(data.data[i])
                    indiv_data[i] = np.mean(\
                        [np.cov(
                            x_i[trial,:,~np.isnan(x_i[trial,0,:])].T)\
                            for trial in range(x_i.shape[0]) 
                            if ~np.isnan(x_i[trial,0,:]).all()],axis=0)
                pca_ready_data = np.mean(np.array(indiv_data),axis=0)
            else:#assumes ERPs
                if averaged:
                    erps = []
                    for part in data.participant:
                        erps.append(data.sel(participant=part).groupby('samples').mean('epochs').T)
                    pca_ready_data = np.nanmean(erps,axis=0)
                else: 
                    pca_ready_data = data.stack({'all':['participant','epochs','samples']}).dropna('all')
                    pca_ready_data = pca_ready_data.transpose('all','channels')
            # Performing spatial PCA on the average var-cov matrix
            pca_weights = _pca(pca_ready_data, n_comp, data.coords["channels"].values)
            data = data @ pca_weights
            data.attrs['pca_weights'] = pca_weights
    elif method == 'mcca':
        ori_coords = data.drop_vars('channels').coords
        if n_ppcas is None:
            n_ppcas = n_comp*3
        mcca_m = mcca.MCCA(n_components_pca=n_ppcas, n_components_mcca=n_comp)
        if cov:
            fitted_data = data.transpose('participant','epochs','samples','channels').data
            ccs = mcca_m.obtain_mcca_cov(fitted_data)
        else:
            if averaged:
                fitted_data = data.mean('epochs').transpose('participant','samples','channels').data
            else:
                fitted_data = data.stack({'all':['epochs','samples']})\
                .transpose('participant','all','channels').data
            ccs = mcca_m.obtain_mcca(fitted_data, r=mcca_reg)
        trans_ccs = np.tile(np.nan, (data.sizes['participant'], data.sizes['epochs'], data.sizes['samples'], ccs.shape[-1]))
        for i, part in enumerate(data.participant):
                trans_ccs[i] = mcca_m.transform_trials(data.sel(participant=part).transpose('epochs','samples', 'channels').data.copy())
        data = xr.DataArray(trans_ccs,
             dims = ["participant","epochs","samples","component"],
             coords=dict(
                 participant=data.participant,
                 epochs=data.epochs,
                 samples=data.samples,
                 component = np.arange(n_comp))#n_comp
            )
        data = data.assign_coords(ori_coords)
        data.attrs['mcca_weights'] = mcca_m.mcca_weights
        data.attrs['pca_weights'] = mcca_m.pca_weights
    elif method is None:
        data = data.rename({'channels':'component'})
        data['component'] = np.arange(len(data.component))
        data.attrs['pca_weights'] = np.identity(len(data.component))
    else:
        raise ValueError(f"method {method} is unknown, choose either 'pca', 'mcca' or None")

    if apply_zscore:
        ori_coords = data.coords
        match apply_zscore:
            case 'all':
                if zscore_acrossPCs:
                    data = zscore_xarray(data)
                else:
                    data = data.stack(comp=['component']).groupby('comp').map(zscore_xarray).unstack()
            case 'participant':
                if zscore_acrossPCs:
                    data = data.groupby('participant').map(zscore_xarray)
                else:
                    data = data.stack(participant_comp=[participants_variable,'component']).groupby('participant_comp').map(zscore_xarray).unstack()
            case 'trial':
                if zscore_acrossPCs:
                    data = data.stack(trial=[participants_variable,'epochs']).groupby('trial').map(zscore_xarray).unstack()
                else:
                    data = data.stack(trial=[participants_variable,'epochs','component']).groupby('trial', squeeze=False).map(zscore_xarray).unstack()
        data = data.transpose('participant','epochs','samples','component')
        data = data.assign_coords(ori_coords)

    data.attrs['pca_weights'] = pca_weights
    data.attrs['sfreq'] = sfreq
    data = stack_data(data)
    return data
    

def stage_durations(times):
    '''
    Returns the stage durations from the event onset times by substracting each stage to the previous
    
    Parameters
    ----------
    times : ndarray
        2D ndarray with [n_trials * n_events]
    
    Returns
    -------
    times : ndarray
        2D ndarray with [n_trials * n_events]
    '''
    times = np.diff(times, axis=1, prepend=0)
    return times

def save_fit(data, filename):
    '''
    Save fit
    '''
    data.unstack().to_netcdf(filename)
    print(f"{filename} saved")

def load_fit(filename):
    '''
    Load fit
    '''
    data = xr.open_dataset(filename)
    if 'trials' in data:
        data = data.stack(trial_x_participant=["participant","trials"]).dropna(dim="trial_x_participant", how='all')
    return data

def save_eventprobs(eventprobs, filename):
    '''
    Saves eventprobs to filename csv file
    '''
    eventprobs = eventprobs.unstack()
    eventprobs.to_dataframe().to_csv(filename)
    print(f"Saved at {filename}")

def centered_activity(data, times, channels, event, n_samples=None, center=True, cut_after_event=0, baseline=0, cut_before_event=0, event_width=0, impute=None):
    '''
    Parses the single trial signal of a given channel in a given number of samples before and after an event.

    Parameters
    ----------
    data : xr.Dataset
        HMP data (untransformed but with trial and participant stacked)
    times : xr.DataArray
        Onset times as computed using onset_times()
    channels : list
        channels to pick for the parsing of the signal, must be a list even if only one
    event : int 
        Which event is used to parse the signal 
    n_samples : int
        How many samples to record after the event (default = maximum duration between event and the consecutive event)
    cut_after_event: int
        Which event after ```event``` to cut samples off, if 1 (Default) cut at the next event
    baseline: int
        How much samples should be kept before the event
    cut_before_event: int
        At which previous event to cut samples from, ```baseline``` if 0 (Default), no effect if baseline = 0
    event_width: int
        Duration of the fitted events, used when cut_before_event is True

    Returns
    -------
    centered_data : xr.Dataset
        Xarray dataset with electrode value (data) and trial event time (time) and with trial_x_participant * samples dimension
    '''
    if event == 0:#no samples before stim onset
        baseline = 0
    elif event == 1:#no event at stim onset
        event_width = 0
    if cut_before_event == 0:#avoids searching before stim onset
        cut_before_event = event
    if n_samples is None:
        if cut_after_event is None:
            raise ValueError('One of ```n_samples``` or ```cut_after_event``` has to be filled to use an upper limit')
        n_samples = max(times.sel(event=event+cut_after_event).data- 
                                   times.sel(event=event).data)+1
    if impute is None:
        impute = np.nan
    if center:
        centered_data = np.tile(impute, (len(data.trial_x_participant), len(channels),
            int(round(n_samples-baseline+1))))
    else:
        centered_data = np.tile(impute, (len(data.trial_x_participant), len(channels),
            len(data.samples)))
        
    i = 0
    trial_times = np.zeros(len(data.trial_x_participant))*np.nan
    valid_indices = list(times.groupby('trial_x_participant', squeeze=False).groups.keys())
    for trial, trial_dat in data.groupby('trial_x_participant', squeeze=False):
        if trial in valid_indices:
            if cut_before_event>0:
                #Lower lim is baseline or the last sample of the previous event
                lower_lim = np.max([
                    -np.max([times.sel(event=event, trial_x_participant=trial)-
                            times.sel(event=event-cut_before_event, trial_x_participant=trial)-
                        event_width//2,0]), baseline])
            else:
                lower_lim = 0
            if cut_after_event>0:
                upper_lim = np.max([np.min([times.sel(event=event+cut_after_event, trial_x_participant=trial) - times.sel(event=event, trial_x_participant=trial)- event_width//2, \
                                            n_samples]), 0])
            else:
                upper_lim = n_samples
            
            # Determine samples in the signal to store
            start_idx = int(times.sel(event=event, trial_x_participant=trial) + lower_lim)
            end_idx = int(times.sel(event=event, trial_x_participant=trial) + upper_lim)
            trial_time = slice(start_idx, end_idx)
            trial_time_idx =  slice(start_idx, end_idx+1)
            trial_elec = trial_dat.sel(channels = channels, samples=trial_time).squeeze('trial_x_participant')
            # If center, adjust to always center on the same sample if lower_lim < baseline
            baseline_adjusted_start = int(abs(baseline - lower_lim))
            baseline_adjusted_end = baseline_adjusted_start + trial_elec.shape[-1]
            trial_time_arr = slice(baseline_adjusted_start, baseline_adjusted_end)

            if center:
                centered_data[i, :,  trial_time_arr] = trial_elec 
            else:
                centered_data[i, :, trial_time_idx] = trial_elec
            trial_times[i] = times.sel(event=event, trial_x_participant=trial)
            i += 1
    
    part, trial = data.coords['participant'].values, data.coords['epochs'].values
    trial_x_part = xr.Coordinates.from_pandas_multiindex(MultiIndex.from_arrays([part,trial],\
              names=('participant','trials')),'trial_x_participant')
    centered_data = xr.Dataset({'data': (('trial_x_participant','channel','samples'), centered_data),
                          'times': (('trial_x_participant'), trial_times)},
                            {'channel':channels,
                            'samples':np.arange(centered_data.shape[-1])+baseline},
                          
                        attrs = {
                                 'event':event})

    return centered_data.assign_coords(trial_x_part)
    
def condition_selection(hmp_data, epoch_data, condition_string, variable='event', method='equal'):
    '''
    condition_selection select a subset from hmp_data. It selects epochs for which
    'condition_string' is in 'variable' based on 'method'.

    Parameters
    ----------
    hmp_data : xr.Dataset
        transformed EEG data for hmp, from utils.transform_data
    epoch_data : deprecated
    condition_string : str | num
        condition indicator for selection
    variable : str
        variable present in hmp_data that is used for condition selection
    method : str
        'equal' selects equal trials, 'contains' selects trial in which conditions_string
        appears in variable

    Returns
    -------
    dat : xr.Dataset
        Subset of hmp_data.
        
    '''
    unstacked = hmp_data.unstack()
    unstacked[variable] = unstacked[variable].fillna("")
    if method == 'equal':
        unstacked = unstacked.where(unstacked[variable] == condition_string, drop=True)
        stacked = stack_data(unstacked)
    elif method == 'contains':
        unstacked = unstacked.where(unstacked[variable].str.contains(condition_string),drop=True)
        stacked = stack_data(unstacked)
    else:
        print('unknown method, returning original data')
        stacked = hmp_data
    return stacked

def condition_selection_epoch(epoch_data, condition_string, variable='event', method='equal'):

    if len(epoch_data.dims) == 4:
        stacked_epoch_data = epoch_data.stack(trial_x_participant=('participant','epochs')).dropna('trial_x_participant',how='all')

    if method == 'equal':
        stacked_epoch_data = stacked_epoch_data.where(stacked_epoch_data[variable] == condition_string, drop=True)
    elif method == 'contains':
        stacked_epoch_data = stacked_epoch_data.where(stacked_epoch_data[variable].str.contains(condition_string),drop=True)
    return stacked_epoch_data.unstack()

def load_data(path):
    return xr.load_dataset(path)

    
def participant_selection(hmp_data, participant):
    unstacked = hmp_data.unstack().sel(participant = participant)
    stacked = stack_data(unstacked)
    return stacked

def filter_non_converged(estimates):
    for iteration in estimates.iteration.values:
        if np.diff(estimates.sel(iteration=iteration).traces.dropna('em_iteration')[-2:]) < -1e-10:
            estimates = estimates.drop_sel({'iteration':iteration})
    estimates["iteration"] = range(len(estimates.iteration))
    return estimates


#Save eeg/meg data with separate event file, and event_ids
def save_raw_events(fname, data, events, event_dict=None, overwrite=False):
    '''
    Saves raw MNE EEG/MEG data with a separate event file and event_ids, as:
    fname.fif (data)
    fname-eve.fif (events)
    fname-eve.txt (event_ids)

    Parameters
    ----------
    fname : str
        file name, following convention should end in eeg.fif
    data : mne.io.Raw
        EEG/MEG data object
    events : array of int, shape (n_events, 3)
        The array of events. The first column contains the event time in samples, 
        with first_samp included. The third column contains the event id.
    event_dict : dict
        Dictionary of {str: int} mappings of event IDs.
    overwrite : bool
        If True (default False), overwrite the destination file if it exists.
    '''

    #save data
    data.save(fname, overwrite=overwrite)
    
    #save events, same fname + -eve
    ftype = fname[-4:]
    fname_events = fname[:-4] + '-eve' + ftype
    mne.write_events(fname_events, events, overwrite=overwrite)
    
    #save event_dict if given
    if event_dict:
        with open(fname_events[:-4] + '.txt', 'w') as fp:
            json.dump(event_dict, fp)


#Load eeg/meeg data with event file and event_ids if exist
def load_raw_events(fname,preload=False):
    '''
    Load eeg/meeg data with event file and event_ids if exist, saved with save_raw_events

    Parameters
    ----------
    fname : str
        file name, following convention should end in eeg.fif, event file is then assumed
        to end in eeg-eve.fif, event dict in eeg-eve.txt
    preload : bool
        Preload data into memory for data manipulation and faster indexing. If True, the 
        data will be preloaded into memory (fast, requires large amount of memory). 
        If preload is a string, preload is the file name of a memory-mapped file which 
        is used to store the data on the hard drive (slower, requires less memory).

    Returns
    -------
    raw : mne.io.Raw
        Raw object with EEG/MEG data
    events : array of int, shape (n_events, 3)
        The array of events. The first column contains the event time in samples, with 
        first_samp included. The third column contains the event id.
    event_id : dict
        Dictionary of {str: int} mappings of event IDs.
    '''
    #load data
    raw = mne.io.read_raw(fname, preload=preload)

    #load events
    ftype = fname[-4:]
    fname_events = fname[:-4] + '-eve' + ftype
    if os.path.exists(fname_events):
        events = mne.read_events(fname_events)
    else:
        warnings.warn('No event file, data still loaded.')
        events = None

    #load event_dict
    fname_event_dict = fname_events[:-4] + '.txt'
    if os.path.exists(fname_event_dict):
        with open(fname_event_dict, 'r') as fp:
            event_dict = json.load(fp)
    else:
        warnings.warn('No event dictionary found.')
        event_dict = None

    return raw, events, event_dict


def epoch_between_events(raw, events, event_id_from, event_id_to, baseline=None, picks=None, reject=None, tmin=0, tmax=0, decim=1, reject_by_annotation=True, proj='delayed', metadata=None, resample_freq = None, verbose=None):
    '''
    Epoch data between a 'from event' and a 'to event', typically between stimulus and response.
    Empty samples due to different epoch lengths are filled with nans. There must be the same 
    number of from and to events.

    NOTE: reject and resample of resulting Epochs object do not work properly!

    Parameters
    ----------
    raw : mne.io.Raw object
        An instance of Raw.
    events : array of int, shape (n_events, 3)
        The array of events. The first column contains the event time in samples, with first_samp included. 
        The third column contains the event id.
    event_id_from : dict
        The id of the 'from events' to consider. The keys can later be used to access associated events. Example: dict(auditory=1, visual=3). 
    event_id_to : dict
        The id of the 'to events' to consider. The keys can later be used to access associated events. Example: dict(correct=11, incorrect=13).
    baseline : None | tuple of length 2
        The time interval to consider as “baseline” with respect to the 'from events' when applying baseline correction. If None, do not apply baseline correction. If a tuple (a, b), the interval is between a and b (in seconds), including the endpoints. If a is None, the beginning of the data is used; and if b is None, it is set to the end of the interval. If (None, None), the entire time interval is used.
    picks : str | array_like | slice | None
        Channels to include. Slices and lists of integers will be interpreted as channel indices. In lists, channel type strings (e.g., ['meg', 'eeg']) will pick channels of those types, channel name strings (e.g., ['MEG0111', 'MEG2623'] will pick the given channels. Can also be the string values “all” to pick all channels, or “data” to pick data channels. None (default) will pick all channels. Note that channels in info['bads'] will be included if their names or indices are explicitly provided.
    reject : dict | None
        Reject epochs based on maximum peak-to-peak signal amplitude (PTP), i.e. the absolute difference between the lowest and the highest signal value. In each individual epoch, the PTP is calculated for every channel. If the PTP of any one channel exceeds the rejection threshold, the respective epoch will be dropped. The dictionary keys correspond to the different channel types; valid keys can be any channel type present in the object. Example:
        reject = dict(grad=4000e-13,  # unit: T / m (gradiometers)
              mag=4e-12,      # unit: T (magnetometers)
              eeg=40e-6,      # unit: V (EEG channels)
              eog=250e-6      # unit: V (EOG channels)
              )
    tmin, tmax : float
        Start and end time of the epochs in seconds, relative to respectively the from and the end events. The closest or matching samples corresponding to the start and end time are included. Defaults to 0.
    decim : int
        Factor by which to subsample the data.
    reject_by_annotation : bool
        Whether to reject based on annotations. If True (default), epochs overlapping with segments whose description begins with 'bad' are rejected. If False, no rejection based on annotations is performed.
    proj : bool | ‘delayed’
        Apply SSP projection vectors. If proj is ‘delayed’ and reject is not None the single epochs will be projected before the rejection decision, but used in unprojected state if they are kept. This way deciding which projection vectors are good can be postponed to the evoked stage without resulting in lower epoch counts and without producing results different from early SSP application given comparable parameters. Note that in this case baselining, detrending and temporal decimation will be postponed. If proj is False no projections will be applied which is the recommended value if SSPs are not used for cleaning the data.    
    metadata : instance of pandas.DataFrame | None
        A pandas.DataFrame specifying metadata about each epoch. If given, len(metadata) must equal len(events). The DataFrame may only contain values of type (str | int | float | bool). If metadata is given, then pandas-style queries may be used to select subsets of data, see mne.Epochs.__getitem__(). When a subset of the epochs is created in this (or any other supported) manner, the metadata object is subsetted accordingly, and the row indices will be modified to match epochs.selection.
    resample_freq: float | None
        Resample data to resample_freq.
     verbose : bool | str | int | None
        Control verbosity of the logging output. If None, use the default verbosity level. See the logging documentation and mne.verbose() for details. Should only be passed as a keyword argument.   

    Returns
    -------
    epochs : mne.Epochs
        mne.Epochs object with epoched data, missing samples are np.nan
    '''

    #set params for raw.get_data()
    reject_by_annotation = 'NaN' if reject_by_annotation else None

    #get data and fill matrix
    from_id = list(event_id_from.values())
    to_id = list(event_id_to.values())

    n_epochs = np.count_nonzero(np.isin(events[:,2],from_id))
    n_channels = len(raw.get_channel_types(picks=picks))

    epochs_start_stop = np.zeros((n_epochs,4), dtype='int64') #start and stop samples, length, and event from id 

    #find start ids in events and following to id, store
    epochs_cnt = 0
    for i, ev in enumerate(events):
        if ev[2] in from_id:
            end_ev = i
            while events[end_ev,2] not in to_id:
                end_ev += 1
                if end_ev == len(events):
                    raise ValueError(
                        "Every from_id must have a following to_id"
                    )
            epochs_start_stop[epochs_cnt] = [ev[0], events[end_ev,0], 0, ev[2]]
            epochs_cnt += 1

    #apply tmin, tmax; calc len
    epochs_start_stop[:,0] = epochs_start_stop[:,0] + round(tmin * raw.info["sfreq"])
    epochs_start_stop[:,1] = epochs_start_stop[:,1] + round(tmax * raw.info["sfreq"])
    epochs_start_stop[:,2] = epochs_start_stop[:,1] - epochs_start_stop[:,0]
    if any(epochs_start_stop[:,2] < 0):
        raise ValueError("Negative epoch length due to tmin/tmax application.")
    
    #make data matrix, epochs x channels x timepoint (e.g. 228, 25, 500)
    nr_samples = max(epochs_start_stop[:,2])
    dat = np.empty((n_epochs, n_channels, nr_samples))
    dat[:] = np.nan
    drop_log = [()] * n_epochs

    #fill matrix
    for i, ep in enumerate(epochs_start_stop):

        #get data from raw, taking annotations into account
        dat[i,:,:ep[2]] = raw.get_data(picks=picks, start = ep[0], stop = ep[1], reject_by_annotation=reject_by_annotation, verbose=verbose)

        #check reject:
        #missing data due to annotations:
        if np.isnan(dat[i,:,:ep[2]]).any():
            drop_log[i] = drop_log[i] + ('BAD_artifact',)

        #reject based on range
        if reject != None:
            if not np.isnan(dat[i,:,:ep[2]]).all():
                for k,v in reject.items():
                    ch_types = raw.get_channel_types(picks=picks)
                    for ch in np.where(np.array(ch_types) == k)[0]:
                        if np.nanmax((dat[i,ch,:ep[2]])) - np.nanmin((dat[i,ch,:ep[2]])) > v:
                            drop_log[i] = drop_log[i] + (raw.ch_names[ch],)
    
    #remove drops  
    dat = np.delete(dat, [True if tmp != () else False for tmp in drop_log], 0)
    events = mne.pick_events(events, include=from_id)
    events_drop = np.delete(events, [True if tmp != () else False for tmp in drop_log], 0)

    #resample
    #based on Epochs.resample, but taking care of NaNs
    if resample_freq != None:
        
        sfreq = float(resample_freq)
        o_sfreq = raw.info["sfreq"]

        nr_samples = int(round(dat.shape[2] * sfreq/o_sfreq)) #perhaps + 1 to make sure
        dat_resampled = np.empty((dat.shape[0], dat.shape[1], nr_samples))
        dat_resampled[:] = np.nan

        #resample by epoch because NaNs need to be removed
        for i, ep in enumerate(dat):
            #filter non-nan samples
            ep_resamp = mne.filter.resample(ep[:, ~np.isnan(ep[0,:])], sfreq, o_sfreq, npad="auto", window="boxcar", pad="edge")
            dat_resampled[i,:,:ep_resamp.shape[1]] = ep_resamp

        #check lowpass
        lowpass = raw.info.get("lowpass")
        lowpass = np.inf if lowpass is None else lowpass
        with raw.info._unlock():
            raw.info["lowpass"] = min(lowpass, sfreq / 2.0)
            raw.info["sfreq"] = float(sfreq)
        
        dat = dat_resampled

    #create epochs object
    # NOTE using reject here does not work, it only rejects (if necessary) the first episode.
    # This seems due to the NaNs.
    epochs = mne.EpochsArray(dat, raw.info, events = events_drop, tmin=tmin, event_id=event_id_from, baseline=baseline, proj=proj, metadata=metadata, drop_log=tuple(drop_log), verbose=verbose)

    #decimate if asked for
    if decim > 1:
        epochs = epochs.decimate(decim, offset=0, verbose=verbose)

    return epochs



'''

'''
import scipy.stats as stats
import scipy.signal as ssignal
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from itertools import cycle
import warnings
default_colors =  ['cornflowerblue','indianred','orange','darkblue','darkgreen','gold']


def plot_topo_timecourse(channels, estimated, channel_position, init, time_step=1, ydim=None,
                figsize=None, dpi=100, magnify=1, times_to_display=None, cmap='Spectral_r',
                ylabels=[], xlabel = None, max_time = None, vmin=None, vmax=None, title=False, ax=None, 
                sensors=False, skip_channels_computation=False, contours=6, event_lines='tab:orange',
                colorbar=True, topo_size_scaling=False, as_time=False,
                linecolors='black',center_measure='mean',estimate_method=None):
    '''
    Plotting the event topologies at the average time of the onset of the next stage.
    
    Parameters
    ----------
    channels : ndarray | xr.Dataarray 
        a 2D or 3D matrix of channel activity with channels and event as dimension (+ eventually a varying dimension) OR
        the original EEG data in HMP format
    estimated : ndarray | hmp object
        a 1D or 2D matrix of times with event as dimension OR directly the results from a fitted hmp (either from fit_single or fit_single_conds)
    channel_position : ndarray
        Either a 2D array with dimension channel and [x,y] storing channel location in meters or an info object from
        the mne package containning digit. points for channel location
    init : float
        initialized HMP object
    time_step : float
        What unit to multiply all the times with, if you want to go on the second or millisecond scale you can provide 
        1/sf or 1000/sf where sf is the sampling frequency of the data
    ydim: str
        name for the extra dimensions (e.g. iteration)
    figsize : list | tuple | ndarray
        Length and heigth of the matplotlib plot
    dpi : float
        DPI of the  matplotlib plot
    magnify : float
        How much should the events be enlarged, useful to zoom on topologies, providing any other value than 1 will 
        however change the displayed size of the event
    times_to_display : ndarray
        Times to display (e.g. Reaction time or any other relevant time) in the time unit of the fitted data
    cmap : str
        Colormap of matplotlib
    xlabel : str
        label of x-axis, default = None, which give "Time (samples)" or "Time" in case time_step != 1
    ylabels : dict
        dictonary with {label_name : label_values}. E.g. {'Condition': ['Speed','Accuracy']}
    max_time : float
        limit of the x (time) axe
    vmin : float
        Colormap limits to use (see https://mne.tools/dev/generated/mne.viz.plot_topomap.html). If not explicitly
        set, uses min across all topos while keeping colormap symmetric.
    vmax : float
        Colormap limits to use (see https://mne.tools/dev/generated/mne.viz.plot_topomap.html). If not explicitly
        set, uses max across all topos while keeping colormap symmetric.
    title : str
        title of the plot
    ax : matplotlib.pyplot.ax
        Matplotlib object on which to draw the plot, can be useful if you want to control specific aspects of the plots
        outside of this function
    sensors : bool
        Whether to plot the sensors on the topologies
    skip_channel_contribution: bool
        if True assumes that the provided channel argument is already topologies for each channel
    contours : int / array_like
        The number of contour lines to draw (see https://mne.tools/dev/generated/mne.viz.plot_topomap.html)
    event_lines : bool / color
        Whether to plot lines and shading to indicate the moment of the event. If True uses tab:orange, if
        set as color, uses the color
    colorbar : bool
        Whether a colorbar is plotted.
    topo_size_scaling : bool
        Whether to scale the size of the topologies with the event size. If True, size of topologies depends
        on total plotted time interval, if False it is only dependent on magnify.
    as_time : bool
        if true, plot time (ms) instead of samples (time_step takes precedence). Ignored if times are provided as array.
    center_measure : string
        mean (default) or median, used to calculate the time within participant
    estimate_method : string
        'max' or 'mean', either take the max probability of each event on each trial, or the weighted 
        average.
    Returns
    -------
    ax : matplotlib.pyplot.ax
        if ax was specified otherwise returns the plot
    '''

    from mne.viz import plot_brain_colorbar, plot_topomap
    from mne import Info
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes

    if as_time and time_step != 1:
        print('time_step takes precedence over as_time')
        as_time = False
    
    cond_plot = False

    #if condition estimates, prep conds
    if isinstance(estimated, xr.Dataset) and 'condition' in estimated.dims:
        cond_plot = True
        conds = estimated['cond'].values
        n_cond = estimated.parameters.shape[0]

        #make times_to_display in list with lines per condition
        default = init.compute_times(init, estimated, mean=True, add_rt=True, extra_dim='condition', center_measure=center_measure,estimate_method=estimate_method).values[:,-1].tolist() #compute corresponding times
        default.reverse()
        if times_to_display is None:
            times_to_display = default
        elif isinstance(times_to_display, list):
            if not isinstance(times_to_display[0], list): #assume it's a list that needs to be copied
                times_to_display = times_to_display * n_cond
            elif len(times_to_display) != n_cond:
                print('times_to_display should either be a list of length n_cond or an ndarray which will be repeated across conditions')
                times_to_display = default
        elif isinstance(times_to_display, np.ndarray): #only one set of times
            if len(times_to_display.shape) == 1:
                times_to_display = [times_to_display] * n_cond

        #set ylabels to conditions
        if ylabels == []:
            ylabels = estimated.clabels

    return_ax = True

    #if not times specified, plot average RT
    if times_to_display is None:
        times_to_display = init.compute_times(init, estimated, mean=True, add_rt=True, extra_dim=None, center_measure=center_measure,estimate_method=estimate_method).values[-1]
        if not isinstance(times_to_display, float):
            times_to_display = times_to_display[-1]

    #set xlabel depending on time_step
    if xlabel is None:
        if time_step == 1 and as_time == False:
            xlabel = 'Time (in samples)'
        elif as_time:
            xlabel = 'Time (ms)'
        else:
            xlabel = 'Time'

    #set color of event_lines 
    if event_lines == True:
        event_color='tab:orange'
    else:
        event_color=event_lines

    #if estimated is an fitted HMP instance, calculate topos and times 
    if isinstance(estimated, (xr.DataArray, xr.Dataset)) and 'event' in estimated:
        if ydim is None:
            if 'n_events' in estimated.dims and estimated.n_events.count() > 1: #and there are multiple different fits (eg backward estimation)
                ydim = 'n_events' #set ydim to 'n_events'
            elif 'condition' in estimated.dims:
                ydim = 'condition'
        if not skip_channels_computation:
            channels = init.compute_topologies(channels, estimated, init, ydim, estimate_method=estimate_method).data #compute topologies
        times = init.compute_times(init, estimated, mean=True, extra_dim=ydim, as_time=as_time, center_measure=center_measure,estimate_method=estimate_method).data #compute corresponding times

    else:#assumes times/topologies already computed
        times = estimated 
    times = times * time_step
    if len(np.shape(channels)) == 2:
        channels = channels[np.newaxis]
    
    if cond_plot: #reverse order, to make correspond to condition maps
        channels = np.flipud(channels)
        times = np.flipud(times)

    n_iter = np.shape(channels)[0]

    if n_iter == 1:
        times = [times]
    times = np.array(times)

    if as_time:
        time_step = 1000/init.sfreq #time_step still needed below
    event_size = init.event_width_samples * time_step
    
    #based the size of the topologies on event_size and magnify or only on magnify
    if topo_size_scaling: #does topo width scale with time interval of plot?
        topo_size = event_size * magnify
    else:
        timescale = (max_time if max_time else (np.max(times_to_display) * time_step * 1.05)) + time_step
        topo_size = .08 * timescale * magnify #8% of time scale

    #fix vmin/vmax across topos, while keeping symmetric
    if vmax == None: #vmax = absolute max, unless no positive values
        vmax = np.nanmax(np.abs(channels[:])) 
        vmin = -vmax if np.nanmin(channels[:]) < 0 else 0
        if np.nanmax(channels[:]) < 0: vmax = 0

    #make axis
    if ax is None:
        if figsize == None:
            if n_iter == 1:
                figsize = (12, .7 * np.max([magnify,1.8])) #make sure they don't get too flat
            else:
                figsize = (12, n_iter*.7) #make sure they don't get too flat

        _, ax = plt.subplots(1, 1, figsize=figsize, dpi=dpi)
        return_ax = False

    axes = []

    #plot row by row
    rowheight = 1/n_iter 
    for iteration in np.arange(n_iter):
        times_iteration = times[iteration]
        missing_evts = np.where(np.isnan(times_iteration))[0]
        times_iteration = np.delete(times_iteration,missing_evts)
        channels_ = channels[iteration,:,:]
        channels_ = np.delete(channels_, missing_evts, axis=0)
        n_event = len(times_iteration)
        ylow = iteration * rowheight

        #plot topology per event
        for event in np.arange(n_event):

            #topology
            axes.append(ax.inset_axes([times_iteration[event]-topo_size/2, ylow+.1*rowheight,
                                topo_size, rowheight*.8], transform=ax.get_xaxis_transform())) 
            plot_topomap(channels_[event,:], channel_position, axes=axes[-1], show=False,
                         cmap=cmap, vlim=(vmin, vmax), sensors=sensors, contours=contours)

            #lines/fill of detected event
            if event_lines:
                #bottom of row + 5% if n_iter > 1
                ylow2 = iteration * rowheight if n_iter == 1 else iteration * rowheight + .05 * rowheight
                #top of row - 5% if n_iter > 1
                yhigh = (iteration + 1) * rowheight if n_iter == 1 else (iteration + 1) * rowheight - .05 * rowheight

                ax.vlines(times_iteration[event]-event_size/2,ylow2,yhigh, linestyles='dotted',color=event_color,alpha=.5,transform=ax.get_xaxis_transform())
                ax.vlines(times_iteration[event]+event_size/2,ylow2, yhigh, linestyles='dotted',color=event_color,alpha=.5, transform=ax.get_xaxis_transform())
                ax.fill_between(np.array([times_iteration[event]-event_size/2,times_iteration[event]+event_size/2]), ylow2, yhigh, alpha=0.15,color=event_color, transform=ax.get_xaxis_transform(),edgecolor=None)

        #add lines per condition
        if cond_plot and times_to_display is not None:
            #bottom of row + 5% if n_iter > 1
            ylow = iteration * rowheight if n_iter == 1 else iteration * rowheight + .05 * rowheight
            #top of row - 5% if n_iter > 1
            yhigh = (iteration + 1) * rowheight if n_iter == 1 else (iteration + 1) * rowheight - .05 * rowheight
            ax.vlines(times_to_display[iteration] * time_step,ylow,yhigh, linestyles='--',transform=ax.get_xaxis_transform())

    #legend
    if colorbar:
        cheight = 1 if n_iter == 1 else 2/n_iter 
        #axins = ax.inset_axes(width="0.5%", height=cheight, loc="lower left", bbox_to_anchor=(1.025, 0, 2, 1), bbox_transform=ax.transAxes, borderpad=0)
        axins = ax.inset_axes([1.025, 0, .03,cheight])
        if isinstance(channel_position, Info):
            lab = 'Voltage' if channel_position['chs'][0]['unit'] == 107 else channel_position['chs'][0]['unit']._name
        else:
            lab = 'Voltage'
        plot_brain_colorbar(axins, dict(kind='value', lims = [vmin,0,vmax]),colormap=cmap, label = lab, bgcolor='.5', transparent=None)

    #plot ylabels
    if isinstance(ylabels, dict):
        tick_labels = [str(x) for x in list(ylabels.values())[0]]
        if cond_plot: tick_labels.reverse() 
        ax.set_yticks(np.arange(len(list(ylabels.values())[0]))+.5,
                      tick_labels)
        ax.set_ylabel(str(list(ylabels.keys())[0]))
    else:
        ax.set_yticks([])
        ax.spines['left'].set_visible(False)
   
    #add vlines across all rows
    if not cond_plot:
        __display_times(ax, times_to_display, 0, time_step, max_time, times, linecolors, n_iter)
    else:
        ax.set_xlim(-1*time_step, np.max(times_to_display)*time_step * 1.05)
    if not return_ax:
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_ylim(0, n_iter) #-1
        ax.set_xlabel(xlabel)
        if title:
            ax.set_title(title)
        if np.any(max_time) == None and np.any(times_to_display) == None:
            ax.set_xlim(0, np.nanmax(times)+np.nanmax(times)/10)
    if plt.get_backend()[0:2] == 'Qt' or plt.get_backend() == 'nbAgg': #fixes issue with yscaling
    #    plt.tight_layout()
        plt.gcf().subplots_adjust(top=0.85,bottom=.2)   #tight layout didn't work anymore
    if return_ax:
        ax.set_ylim(0, n_iter) #-1
        return ax
    else:
        plt.show()    

def save_model_topos(channels, estimated, channel_position, init, fname='topo', figsize=None, dpi=300, cmap='Spectral_r',
                vmin=None, vmax=None, sensors=False, contours=6, colorbar=True):
    '''
    Saving the event topologies to files, one per topology. Typically used for saving high quality topos.
    
    Parameters
    ----------
    channels : xr.Dataarray 
       the original EEG data in HMP format
    estimated : hmp object
        a fitted hmp (either from fit_single, fit_single_conds, or backward_estimation)
    channel_position : ndarray
        Either a 2D array with dimension channel and [x,y] storing channel location in meters or an info object from
        the mne package containning digit. points for channel location
    init : float
        initialized HMP object
    fname : str
        file name, other model specifications will be appended. Should not include extension.
    figsize : list | tuple | ndarray
        Length and heigth of the matplotlib plot
    dpi : float
        DPI of the  matplotlib plot
    cmap : str
        Colormap of matplotlib
    vmin : float
        Colormap limits to use (see https://mne.tools/dev/generated/mne.viz.plot_topomap.html). If not explicitly
        set, uses min across all topos while keeping colormap symmetric.
    vmax : float
        Colormap limits to use (see https://mne.tools/dev/generated/mne.viz.plot_topomap.html). If not explicitly
        set, uses max across all topos while keeping colormap symmetric.
    sensors : bool
        Whether to plot the sensors on the topologies
    contours : int / array_like
        The number of contour lines to draw (see https://mne.tools/dev/generated/mne.viz.plot_topomap.html)
    colorbar : bool
        Whether a colorbar is saved in a separate file (fname_colorbar)
      
    '''

    from mne.viz import plot_brain_colorbar, plot_topomap
    from mne.io.meas_info import Info

    plot_type = 'default'
    ydim = None

    #if condition estimates, prep conds
    if isinstance(estimated, xr.Dataset) and 'condition' in estimated.dims:
        plot_type = 'cond'
        conds = estimated['cond'].values
        n_cond = estimated.parameters.shape[0]
        cond_names = estimated.clabels
        ydim = 'condition'
    elif 'n_events' in estimated.dims and estimated.n_events.count() > 1:
        plot_type = 'backward'
        ydim = 'n_events'

    #calculate topos
    channels = init.compute_topologies(channels, estimated, init, ydim).data #compute topologies
    
    #fix vmin/vmax across topos, while keeping symmetric
    if vmax == None: #vmax = absolute max, unless no positive values
        vmax = np.nanmax(np.abs(channels[:])) 
        vmin = -vmax if np.nanmin(channels[:]) < 0 else 0
        if np.nanmax(channels[:]) < 0: vmax = 0

    #make axis
    if figsize == None:
        figsize = (2,2)

    #make sure it doesn't display plots
    mplint = mpl.is_interactive()
    if mplint:
        plt.ioff()

    #plot model by model
    if plot_type == 'default':

        #remove potential nans
        nans = np.where(np.isnan(np.mean(channels,axis=1)))[0]
        channels = np.delete(channels, nans, axis=0)
        n_event = channels.shape[0]

        for event in range(n_event):
            plot_name = fname + '_ev' + str(event+1) + '.png'
            fig, ax = plt.subplots(figsize=figsize)
            plot_topomap(channels[event,:], channel_position, show=False, cmap=cmap, vlim=(vmin, vmax), sensors=sensors, contours=contours, axes=ax)
            fig.savefig(plot_name,dpi=dpi,transparent=True)    

    elif plot_type == 'cond': #reverse order, to make correspond to condition maps
        channels = np.flipud(channels)
        cond_labels = [str(x[0]) for x in list(cond_names.values())[0]]
        
        #plot by condition
        for cond in range(n_cond):
            cond_name = cond_labels[cond]
            cond_channels = channels[cond]

            #remove potential nans
            nans = np.where(np.isnan(np.mean(cond_channels,axis=1)))[0]
            cond_channels = np.delete(cond_channels, nans, axis=0)
            n_event = cond_channels.shape[0]

            for event in range(n_event):
                plot_name = fname + '_cond' + cond_name + '_ev' + str(event+1) + '.png'
                fig, ax = plt.subplots(figsize=figsize)
                plot_topomap(cond_channels[event,:], channel_position, show=False, cmap=cmap, vlim=(vmin, vmax), sensors=sensors, contours=contours, axes=ax)
                fig.savefig(plot_name,dpi=dpi,transparent=True)    
    
    elif plot_type == 'backward':

        # plot by number of events
        for n_eve in range(channels.shape[0]):
            back_channels = channels[n_eve]

            #remove potential nans
            nans = np.where(np.isnan(np.mean(back_channels,axis=1)))[0]
            back_channels = np.delete(back_channels, nans, axis=0)
            n_event = back_channels.shape[0]

            for event in range(n_event):
                plot_name = fname + '_backward' + str(n_event) + '_ev' + str(event+1) + '.png'
                fig, ax = plt.subplots(figsize=figsize)
                plot_topomap(back_channels[event,:], channel_position, show=False, cmap=cmap, vlim=(vmin, vmax), sensors=sensors, contours=contours, axes=ax)
                fig.savefig(plot_name,dpi=dpi,transparent=True)    

    #legend
    if colorbar:
        if isinstance(channel_position, Info):
            lab = 'Voltage' if channel_position['chs'][0]['unit'] == 107 else channel_position['chs'][0]['unit']._name
        else:
            lab = 'Voltage'
        fig, ax = plt.subplots(figsize=(.5,2))
        plot_brain_colorbar(ax, dict(kind='value', lims = [vmin,0,vmax]),colormap=cmap, label = lab, bgcolor='.5', transparent=None)
        fig.savefig(fname + '_colorbar.png',dpi=dpi,transparent=True,bbox_inches='tight')    

    #switch plotting back on
    if mplint:
        plt.ion()


def plot_components_sensor(hmp_data, positions):
    """
      This function is used to visualize the topomap of the HMP principal components.
     
     Parameters
     ----------
     	 hmp_data: xr.Dataset
            Data returned from the function hmp.utils.transform_data()
     	 positions: mne.info | ndarray 
            List of x and y positions to plot channels on head model OR MNE info object
    """
    from mne.viz import plot_topomap
    fig, ax = plt.subplots(1,len(hmp_data.attrs['pca_weights'].component))
    for comp in hmp_data.attrs['pca_weights'].component:
        plot_topomap(hmp_data.attrs['pca_weights'].values[:,comp], positions, axes=ax[comp], show=False, cmap='Spectral_r')
    plt.show()


def plot_loocv(loocv_estimates, pvals=True, test='t-test', figsize=(16,5), indiv=True, ax=None, mean=False, additional_points=None, return_pvals=False):
    '''
    Plotting the LOOCV results
    
    Parameters
    ----------
    loocv_estimates : ndarray or xarra.DataArray
        results from a call to hmp.utils.loocv()
    pvals : bool
        Whether to display the pvalue with the associated test
    test : str
        which statistical test to compute for the difference in LOOCV-likelihood (one sample t-test or sign test)
    figsize : list | tuple | ndarray
        Length and heigth of the matplotlib plot
    indiv : bool
        Whether to plot individual lines
    ax : matplotlib.pyplot.ax
        Matplotlib object on which to draw the plot, can be useful if you want to control specific aspects of the plots
        outside of this function
    mean : bool
        Whether to plot the mean
    additional_points : 
        Additional likelihood points to be plotted. Should be provided as a list of tuples 
        containing the x coordinate and loocv estimates with a single event, e.g. [(5,estimates)].
    return_pvals : bool 
        Whether to return the pvalues

    Returns
    -------
    ax : matplotlib.pyplot.ax
        if ax was specified otherwise returns the plot
    '''
    ax=None
    if return_pvals:
        pvals = True
    if pvals:
        if test == 'sign':
            from statsmodels.stats.descriptivestats import sign_test 
        elif test == 't-test':
            from scipy.stats import ttest_1samp
        else:
            raise ValueError('Expected sign or t-test argument to test parameter')
    if ax is None:
        fig, ax = plt.subplots(1,2, figsize=figsize)
        return_ax = False
    else:
        return_ax = True
    loocv_estimates = loocv_estimates.dropna('n_event', how='all')

    #stats
    diffs, diff_bin, labels = [],[],[]
    pvalues = []
    for n_event in np.arange(2,loocv_estimates.n_event.max()+1):
        diffs.append(loocv_estimates.sel(n_event=n_event).data - loocv_estimates.sel(n_event=n_event-1).data) #differences
        diff_bin.append([1 for x in diffs[-1] if x > 0]) #nr of positive differences
        labels.append(str(n_event-1)+'->'+str(n_event))

        if pvals:
            if test == 'sign':
                diff_tmp = np.array(diffs)
                diff_tmp[np.isnan(diff_tmp)] = -np.inf 
                pvalues.append((sign_test(diff_tmp[-1])))
            elif test == 't-test':
                pvalues.append((ttest_1samp(diffs[-1], 0, alternative='greater')))

     
    #first plot
    if mean:
        alpha = .4#for the indiv plot
        marker_indiv = '.'
        means = np.nanmean(loocv_estimates.data,axis=1)[::-1]
        errs = (np.nanstd(loocv_estimates.data,axis=1)/np.sqrt(len(loocv_estimates.participant)))[::-1]
        ax[0].errorbar(x=np.arange(len(means))+1, y=means, \
                 yerr= errs, marker='o', color='k')
        #if pvals:
        #    for n_event in np.arange(2,loocv_estimates.n_event.max()+1):       
        #        mean = means[n_event-1] * 1.1
        #        ax[0].text(x=n_event-.25, y=mean, s=str(int(np.nansum(diff_bin[n_event-2])))+'/'+ str(len(diffs[-1])) + ': ' + str(np.around(pvalues[n_event-2][-1],3)), ha='right')
    else:
        alpha=1
        marker_indiv='o'
    if indiv:
        for loo in loocv_estimates.T:
            ax[0].plot(loocv_estimates.n_event,loo, alpha=alpha,marker=marker_indiv)
    ax[0].set_ylabel('LOOCV Loglikelihood')
    ax[0].set_xlabel('Number of events')
    ax[0].set_xticks(ticks=loocv_estimates.n_event)

    if additional_points: #only plot average for now
        if not isinstance(additional_points,list):
            additional_points = [additional_points]
        for ap in additional_points:
            xap = ap[0]
            meanap = np.mean(ap[1].values)
            err = np.nanstd(ap[1].values)/np.sqrt(len(ap[1].values))
        ax[0].errorbar(x=xap,y=meanap, yerr=err, marker='o')
      
    #second plot
    diffs = np.array(diffs)
    diffs[np.isneginf(diffs)] = np.nan
    diffs[np.isinf(diffs)] = np.nan

    ax[1].plot(diffs,'.-', alpha=.6)
    ax[1].set_xticks(ticks=np.arange(0,loocv_estimates.n_event.max()-1), labels=labels)
    ax[1].hlines(0,0,len(np.arange(2,loocv_estimates.n_event.max())),color='lightgrey',ls='--')
    ax[1].set_ylabel('Change in likelihood')
    ax[1].set_xlabel('')
        
    if pvals:
        ymin = np.nanmin(diffs[:])
        ymintext = ymin - (np.nanmax(diffs[:]) - ymin) * .05
        ymin = ymin - (np.nanmax(diffs[:]) - ymin) * .1
        ax[1].set_ylim(bottom=ymin)
        for n_event in np.arange(2,loocv_estimates.n_event.max()+1):       
            ax[1].text(x=n_event-2, y=ymintext, s=str(int(np.nansum(diff_bin[n_event-2])))+'/'+ str(len(diffs[-1])) + ': ' + str(np.around(pvalues[n_event-2][-1],3)), ha='center')
    
    if return_ax:
        if return_pvals:
            return [ax, pvalues]
        else:
            return ax
    else:
        plt.tight_layout()
        plt.show()
        if return_pvals:
            return pvalues
        

def plot_latencies_average(times, time_step=1, labels=[], colors=default_colors,
    figsize=None, errs='ci', max_time=None, times_to_display=None):
    print('Deprecated. Use plot_latencies() instead.')
    

def plot_distribution(times, colors=default_colors, xlims=False, figsize=(8, 3), survival=False):
    '''
    Plots the distribution of each stage latencies

    Parameters
    ----------
    times : ndarray
        2D or 3D numpy array, Either trials * events or conditions * trials * events
    colors : ndarray
        array of colors for the different stages
    xlims : tuple | list
        lower and upper limit of the x (time) axis
    figsize : list | tuple | ndarray
        Length and heigth of the matplotlib plot
    survival : bool
        Whether to draw density plots or survival density plots

    '''
    f, axs = plt.subplots(1,1, figsize=figsize, dpi=100)

    if len(np.shape(times)) == 2:
        times = np.asarray([times],dtype = 'object')
    cycol = cycle(colors)
    for iteration in times:
        for stage in iteration.T:
            if survival:
                axs.plot(1-stage.cumsum(axis=0), color=next(cycol) )
            else: 
                axs.plot(stage,color=next(cycol) )
    axs.set_ylabel('p(event)')
    axs.set_xlabel('Time (in samples)')
    if xlims:
        axs.set_xlim(xlims[0], xlims[1])
    return axs

def __display_times(ax, times_to_display, yoffset, time_step, max_time, times, linecolors, ymax=1):
    n_iter = len(times)
    times = np.asarray(times,dtype=object)
    if isinstance(times_to_display, (np.ndarray, np.generic)):
        ax.vlines(times_to_display*time_step, yoffset-1.1, yoffset+ymax+1.1, ls='--', colors=linecolors)
        ax.set_xlim(-1*time_step, np.max(times_to_display)*time_step * 1.05)
    if max_time:
        ax.set_xlim(-1*time_step, max_time)
    return ax

def plot_latencies_gamma(gammas, event_width=0, time_step=1, labels=[''], colors=default_colors, 
                         figsize=False, times_to_display=None, max_time=None, kind='bar', legend=False):
    '''
    Plots the average of stage latencies based on the estimated scale parameter of the gamma distributions

    Parameters
    ----------
    gammas : ndarray
        instance of hmp.hmp.parameters
    event_width : float
        Size of the event in time unit given sampling frequency, if drawing a fitted object using hmp you 
        can provide the event_width_sample of fitted hmp (e.g. init.event_width_sample)
    time_step : float
        What unit to multiply all the times with, if you want to go on the second or millisecond scale you can provide 
        1/sf or 1000/sf where sf is the sampling frequency of the data
    labels : tuples | list
        labels to draw on the y axis
    colors : ndarray
        array of colors for the different stages
    figsize : list | tuple | ndarray
        Length and heigth of the matplotlib plot
    times_to_display : ndarray
        Times to display (e.g. Reaction time or any other relevant time) in the time unit of the fitted data
    max_time : float
        limit of the x (time) axe
    kind : str
        Whether to draw a bar plot ('bar') or line connected point plot ('point')
    legend : bool
        Whether to draw legend handle or not
    '''
    j = 0
    
    if len(np.shape(gammas)) == 2:
        gammas = [gammas]#np.reshape([gammas], (1, np.shape(gammas)[0],np.shape(gammas)[1]))
    if not figsize:
        figsize = (8, 1*len(gammas)+2)
    f, axs = plt.subplots(1,1, figsize=figsize, dpi=100)
    for time in gammas:
        cycol = cycle(colors)
        try:
            time = time[np.isfinite(time)]
        except:
            print('todo')
        time = np.reshape(time, (np.shape(gammas)[1],np.shape(gammas)[2]))
        n_stages = int(len(time))
        colors = [next(cycol) for x in np.arange(n_stages)]
        colors.append(next(cycol))
        if kind=='bar':
            axs.bar(np.arange(n_stages)+1, time[:,0] *  time[:,1], color='w', edgecolor=colors)
        elif kind == 'point':
            axs.plot(np.arange(n_stages)+1, time[:,0] * time[:,1],'o-', label=labels[j], color=colors[j])
        j += 1
    #plt.xticks(np.arange(len(labels)),labels)
    #plt.xlim(0-1,j)
    #__display_times(axs, mean_d, np.arange(np.shape(gammas)[0]), time_step, max_time, gammas)
    axs.set_ylabel('Stages durations (Gamma)')
    axs.set_xlabel('Stage number')
    # Hide the right and top spines
    axs.spines.right.set_visible(False)
    axs.spines.top.set_visible(False)
    # Only show ticks on the left and bottom spines
    axs.yaxis.set_ticks_position('left')
    if legend:
        axs.legend()
    return axs

def plot_latencies(times, init=None, time_step=1, labels=[], colors=default_colors,
    figsize=False, errs=None, kind='bar', legend=False, max_time=None, as_time=True):
    '''
    Plots the average of stage latencies with choosen errors bars

    Parameters
    ----------
    times : ndarray | hmp results object
        2D or 3D numpy array, either trials * events or conditions * trials * events
        hmp results object either of fit_single or fit_single_conds
    init : hmp object
        Initialized hmp object, required if times is an hmp results object
    event_width : float
        Display size of the event in time unit given sampling frequency, if drawing a fitted object using hmp you 
        can provide the event_width_sample of fitted hmp (e.g. init.event_width_sample)
    time_step : float
        What unit to multiply all the times with, if you want to go on the second or millisecond scale you can provide 
        1/sf or 1000/sf where sf is the sampling frequency of the data
    labels : tuples | list
        labels to draw on the y axis
    colors : ndarray
        array of colors for the different stages
    figsize : list | tuple | ndarray
        Length and heigth of the matplotlib plot
    errs : str
        Whether to display no error bars (None), 95% confidence interval ('ci'), standard deviation ('std'),
        or standard error ('se')
    times_to_display : ndarray
        Times to display (e.g. Reaction time or any other relevant time) in the time unit of the fitted data
    max_time : float
        limit of the x (time) axe
    kind : str
        bar or point
    as_time : bool
        if true, plot time (ms) instead of samples (time_step takes precedence). Ignored if times are provided as array.
    '''

    from seaborn.algorithms import bootstrap #might be too much to ask for seaborn install?
       
    if as_time and time_step != 1:
        print('time_step takes precedence over as_time')
        as_time = False

    #if hmp estimates are provided, calculate time
    hmp_obj = False
    if isinstance(times, (xr.DataArray, xr.Dataset)):
        assert init is not None, 'If hmp results object provided, init is a required parameter.'
        hmp_obj = True
        ydim = None
        if 'n_events' in times.dims and times.n_events.count() > 1: #and there are multiple different fits (eg backward estimation)
            ydim = 'n_events' 
        elif 'condition' in times.dims:
            ydim = 'condition'
        avg_durations = init.compute_times(init, times, mean=True, duration=True,add_rt=True, extra_dim=ydim, as_time=as_time).data 
        avg_times = init.compute_times(init, times, mean=True, duration=False,add_rt=True, extra_dim=ydim, as_time=as_time).data 
        if errs is not None:
            errorbars = init.compute_times(init, times, duration=True,add_rt=True, extra_dim=ydim, as_time=as_time, errorbars=errs)
        if len(avg_times.shape) == 1:
            avg_durations = np.expand_dims(avg_durations, axis=0)
            avg_times = np.expand_dims(avg_times, axis=0)
            if errs is not None:
                errorbars = np.expand_dims(errorbars, axis=0)

        avg_durations = avg_durations * time_step
        avg_times = avg_times * time_step
        if errs is not None:
            errorbars = errorbars * time_step

        if ydim == 'condition': #reverse order, to make correspond to condition maps
            avg_durations = np.flipud(avg_durations)
            avg_times = np.flipud(avg_times)
            if errs is not None:
                errorbars = np.flipud(errorbars)

        n_model = avg_times.shape[0]
    else: #no hmp object
        if len(np.shape(times)) == 2:
            times = [times]
        times = [time * time_step for time in times]
        if isinstance(times[0], xr.DataArray):
            times = [time.values for time in times]
        n_model = len(times)
        avg_times=[]
        avg_durations=[]
        errorbars =[]
        
    if labels == []:
        if hmp_obj and ydim == 'condition':
            labels = [str(x) for x in reversed(list(times.clabels.values())[0])] 
        else:
            labels = np.arange(n_model)

    if not figsize:
        figsize = (8, 1*n_model+2)
    f, axs = plt.subplots(1,1, figsize=figsize, dpi=100)

    cycol = cycle(colors)
    cur_colors = [next(cycol) for x in np.arange(n_model)] #color per condition/model (line plot)
    
    for j in range(n_model): #per condition/model
        if hmp_obj: #all calcs have been done
            avg_times_model = np.hstack((0,avg_times[j,:]))
            avg_durations_model = avg_durations[j,:]
            n_stages = len(avg_durations_model)
            errorbars_model = errorbars[j,:,:] if errs is not None else None
        else: #call error bars etc
            time = times[j]
            duration = np.diff(time, axis=1, prepend=0) #time per trial, duration per trial
            avg_times_model = np.hstack((0, np.mean(time, axis=0)))
            avg_durations_model = np.mean(duration, axis=0)
            n_stages = len(avg_durations_model)
            errorbars_model = None if errs is None else np.zeros((2,n_stages))
            if errs == 'ci':
                for st in range(n_stages):
                    errorbars_model[:, st] = np.squeeze([np.nanpercentile(bootstrap(duration[:,st]), q=[2.5,97.5])])
                    errorbars_model[:, st] = np.abs(errorbars_model[:, st]-avg_durations_model[st])
            elif errs == 'std':
                errorbars_model = np.tile(np.std(duration, axis=0), (2,1))
            elif errs == 'se':
                print('Cannot calculate SE for non-hmp object')
            avg_durations.append(np.nanmax(avg_durations_model)) #for determining axes
            errorbars.append(np.nanmax(errorbars_model))
            avg_times.append(np.nanmax(avg_times_model))
            
        if kind == 'bar':
            cycol = cycle(colors) #get a color per stage
            cur_colors = [next(cycol) for x in np.arange(n_stages)]

            # #remove 0 stages and associated color
            if np.any(np.isnan(avg_durations_model)):
                missing_evts = np.where(np.isnan(avg_durations_model))[0]
                avg_durations_model = np.delete(avg_durations_model,missing_evts)
                avg_times_model = np.delete(avg_times_model, missing_evts + 1)
                if errs is not None:
                    errorbars_model = np.delete(errorbars_model, missing_evts, axis=1)
                cur_colors = np.delete(cur_colors, missing_evts)
                n_stages = n_stages - len(missing_evts)

            for st in reversed(range(n_stages)): #can't deal with colors in one call
                plt.barh(j, avg_durations_model[st],left=avg_times_model[st], color='w', edgecolor=cur_colors[st])
                if errs is not None:
                    plt.errorbar(np.repeat(avg_times_model[st+1],2), np.repeat(j,2), xerr=errorbars_model[:,st], color=cur_colors[st], fmt='none', capsize=10)

        elif kind == 'point':
            plt.errorbar(np.arange(n_stages)+1, avg_durations_model, 
                    yerr=errorbars_model, color=cur_colors[j], fmt='o-', capsize=10, label=labels[j])
        else:
            raise ValueError('Unknown \'kind\'')
    
    #general settings
    if kind == 'bar':
        plt.yticks(np.arange(len(labels)),labels)
        plt.ylim(0-1,j+1)
        
        if not max_time:
            max_time = (np.nanmax(avg_times) + np.nanmax(errorbars) if errs else np.nanmax(avg_times)) * 1.05               
        axs.set_xlim(0, max_time)
        if time_step == 1 and not as_time:
            plt.xlabel('Cumulative stage durations from stimulus onset (samples)')
        elif as_time:
            plt.xlabel('Cumulative stage durations from stimulus onset (ms)')
        else:
            plt.xlabel('Cumulative stage durations from stimulus onset')
    elif kind == 'point':                 
        plt.xlim(1-.5, n_stages+.5)

        max_y = (np.nanmax(avg_durations) + np.nanmax(errorbars) if errs else np.nanmax(avg_durations)) * 1.05
        axs.set_ylim(0, max_y)

        if time_step == 1 and as_time == False:
            plt.ylabel('Stage durations (samples)')
        elif as_time:
            plt.ylabel('Stage durations (ms)')
        else:
            plt.ylabel('Stage durations')
        plt.xlabel('Stage')
    
    plt.tight_layout()
    # Hide the right and top spines
    axs.spines.right.set_visible(False)
    axs.spines.top.set_visible(False)
    # Only show ticks on the left and bottom spines
    axs.yaxis.set_ticks_position('left')
    axs.xaxis.set_ticks_position('bottom')
    if legend:
        axs.legend()
    return axs


def plot_iterations(iterations, eeg_data, init, positions, dims=['magnitudes','parameters'], alpha=1, ax=None):
    """
    Currently DEPRECATED
    """
    from hmp.models import hmp
    if 'iteration' not in iterations.dims:
        try:
            iterations['iteration'] = [0]
        except:
            iterations = iterations.expand_dims({'iteration':[0]}, axis=1)
            iterations['iteration'] = [0]
    for iteration in iterations.iteration[:-1]:
        selected = init.fit_single(iterations.sel(iteration=iteration)[dims[0]].dropna(dim='event').event[-1].values+1,\
            magnitudes = iterations.sel(iteration=iteration)[dims[0]].dropna(dim='event'),\
            parameters = iterations.sel(iteration=iteration)[dims[1]].dropna(dim='stage'),\
            maximization=False, verbose=False)
        #Visualizing
        plot_topo_timecourse(eeg_data, selected, positions,  init, magnify=1, sensors=False,ax=ax)
    

def plot_bootstrap_results(bootstrapped, info, init, model_to_compare=None, epoch_data=None):
    """
     Plot bootstrapped time courses. This is a wrapper around plot_topo_timecourse to make it easier to use in other functions
     
     Parameters
     ----------
     	 bootstrapped: xr.Dataset 
            The resulting data from a call to hmp.resample.bootstrapping()
     	 info: 
            The MNE info/electrode positions
     	 init: 
            The initialized HMP model
     	 model_to_compare: 
            If None ( default ) the model to compare is taken as the maximal n_event model among the bootstrapped models 
     	 epoch_data: If None ( default
    """
    from hmp.resample import event_occurence
    maxboot_model, labels, counts, event_number, label_event_num = event_occurence(bootstrapped, model_to_compare)
    fig, axes = plt.subplot_mosaic([['a', 'a'], ['b', 'c'], ['b', 'c']],
                              layout='constrained')
    n_events = int(maxboot_model.event.max())
    if model_to_compare is None: 
        plot_topo_timecourse(maxboot_model.channels_activity.values, maxboot_model.event_times.values, info, init,ax=axes['a'],event_lines=False, colorbar=False, times_to_display=None)
        times = maxboot_model.event_times#init.compute_times(init, maxboot_model, mean=True)#computing predicted event times
    else:
        plot_topo_timecourse(epoch_data, model_to_compare, info, init,ax=axes['a'], event_lines=False, colorbar=False, times_to_display=None)
        times = init.compute_times(init, model_to_compare, mean=True)
        maxboot_model = model_to_compare
    counts_adjusted = np.zeros(n_events+1)
    counts_adjusted[:len(counts)] = counts
    for event in times.event.values:
        axes['a'].text(times.sel(event=event).values+init.event_width_samples/2.5, .7, event)
    # axes['a'].set_xlabel('Time (samples)')
    axes['b'].bar(maxboot_model.event,counts_adjusted)
    axes['b'].set_xlabel('Event number')
    axes['b'].set_xticks(maxboot_model.event)
    axes['b'].set_ylabel('Frequency')
    axes['b'].set_ylim(0,1)
    
    axes['c'].bar(label_event_num,event_number)
    axes['c'].set_xlabel('Number of events')
    
    axes['c'].set_ylabel('Frequency')
    axes['c'].set_ylim(0,1)
    axes['a'].set_xticks([])
    axes['a'].spines[['right', 'top', 'bottom']].set_visible(False)
    axes['b'].spines[['right', 'top']].set_visible(False)
    axes['c'].spines[['right', 'top']].set_visible(False)
    # plt.tight_layout()
    plt.show()
    
def plot_expected_distribution(distribution, mean, shape, location=0, xmax=300, xmin=0, num=100, label='distribution', color=None, ax=None, display_mean=False):
    '''
    Plot available distribution with a scale and shape parameter
    '''
    match distribution:
        case 'gamma':
            from scipy.stats import gamma as sp_dist
            from hmp.utils import gamma_scale_to_mean,gamma_mean_to_scale
            scale_to_mean, mean_to_scale = gamma_scale_to_mean,gamma_mean_to_scale
        case 'lognormal':
            from scipy.stats import lognorm as sp_dist
            from hmp.utils import logn_scale_to_mean,logn_mean_to_scale
            scale_to_mean, mean_to_scale = logn_scale_to_mean,logn_mean_to_scale
        case 'wald':
            from scipy.stats import invgauss as sp_dist
            from hmp.utils import wald_scale_to_mean,wald_mean_to_scale
            scale_to_mean, mean_to_scale = wald_scale_to_mean,wald_mean_to_scale
        case 'weibull':
            from scipy.stats import weibull_min as sp_dist
            from hmp.utils import weibull_scale_to_mean,weibull_mean_to_scale
            scale_to_mean, mean_to_scale = weibull_scale_to_mean,weibull_mean_to_scale
        case 'maxwell-boltzmann':
            from scipy.stats import chi as sp_dist
            from hmp.utils import maxb_scale_to_mean,maxb_mean_to_scale
            shape = 3
            scale_to_mean, mean_to_scale = maxb_scale_to_mean,maxb_mean_to_scale
        case 'log-logistic':
            from scipy.stats import fisk as sp_dist
            from hmp.utils import fisk_scale_to_mean,fisk_mean_to_scale
            scale_to_mean, mean_to_scale = fisk_scale_to_mean,fisk_mean_to_scale
        case 'rayleigh':
            from scipy.stats import chi as sp_dist
            from hmp.utils import ray_scale_to_mean,ray_mean_to_scale
            shape = 2
            scale_to_mean, mean_to_scale = ray_scale_to_mean,ray_mean_to_scale
        case 'half-normal':
            from scipy.stats import chi as sp_dist
            from hmp.utils import halfn_scale_to_mean,halfn_mean_to_scale
            shape = 1
            scale_to_mean, mean_to_scale = halfn_scale_to_mean,halfn_mean_to_scale
        case 'uniform':
            from scipy.stats import uniform as sp_dist
            from hmp.utils import uniform_scale_to_mean,uniform_mean_to_scale
            scale_to_mean, mean_to_scale = uniform_scale_to_mean,uniform_mean_to_scale
            location=None
        case _:
                raise ValueError(f'Unknown Distribution {distribution}')
    if ax is None:
        ax = plt
    x = np.linspace(xmin, xmax, num=num)
    if distribution != 'uniform': 
        y = sp_dist.cdf(x, shape, scale=mean_to_scale(mean, shape), loc=location)
    else:
        y = sp_dist.cdf(x, shape, scale=mean_to_scale(mean, shape))
    y = np.diff(y, prepend=0)#going to pmf
    ax.plot(x, y, label=label, color=color)
    if display_mean:
        if distribution != 'uniform': 
            ax.vlines(sp_dist.mean(shape, scale=mean_to_scale(mean, shape), loc=location), np.min(y), np.max(y), color=color)
        else: 
            ax.vlines(sp_dist.mean(shape, scale=mean_to_scale(mean, shape)), np.min(y), np.max(y), color=color)

def plot_estimate_development(estimates, init, epoch_data, info, print_correlations=False, estimate_method='mean'):
    '''
    Plot the shift in estimated event locations over time, with color-coded locations
    for estimated model(s).
     
    Parameters
     ----------
     	estimates : estimated hmp model, or a list of models
            estimated model(s) to be plotted, can be from fit_single, backward estimation, or fit
        init : initialized hmp model
        epoch_data : xr.Dataarray 
            the original EEG data in HMP format
        info : ndarray
            Either a 2D array with dimension channel and [x,y] storing channel location in meters or an info object from
            the mne package containning digit. points for channel location
        print_correlations: bool
            whether to print the correlations between events
        estimate_method : 'max' or 'mean'
            if mean, uses estimates that are used during estimation
            if max, makes plots and correlation based on max (default plotting)
    '''
    
    if not isinstance(estimates, list):
        estimates = [estimates]
    
    #split backward if present
    all_estimates = []
    for est in estimates:
        if not 'n_events' in est.dims:
            all_estimates.append(est)
        else:
            for ev in est.n_events:
                all_estimates.append(est.sel(n_events=ev).dropna("stage", how='all'))

    time_step = 1000/init.sfreq
    
    #get topos
    topos = []
    for est in all_estimates:
        topos.append(init.compute_topologies(epoch_data,est,init,estimate_method=estimate_method).values)
    vm = np.nanmax([np.nanmax(np.abs(x)) for x in topos])

    for est_idx, est in enumerate(all_estimates):

        figsize = (10, 5) 
        plt.subplots(1, 1, figsize=figsize)
    
        #plot development of positions
        neve = len(est.locations) - 1 
        nr = est.traces.shape[0] #nr of estimations

        #plot params
        for i in range(nr):
            cols = np.repeat('black', neve).astype('<U20')
            
            cols[est.locations_dev[i,1:].values == 1] = 'orange'
            cols[est.locations_dev[i,1:].values == 2] = 'gold'
            cols[est.locations_dev[i,1:].values == 3] = 'yellowgreen'
            cols[est.locations_dev[i,1:].values == 4] = 'lawngreen'
            cols[est.locations_dev[i,1:].values == 5] = 'deepskyblue'
            cols[est.locations_dev[i,1:].values == 6] = 'darkblue'
            cols[est.locations_dev[i,1:].values > 6] = 'red'
            plt.scatter(np.cumsum(est.param_dev[i,:,1].values * init.shape * time_step)[:-1],np.repeat(i,neve),color=cols)

        plt.gca().invert_yaxis()
        plt.xlim(0,init.mean_d*time_step)

        #plot topos
        plot_topo_timecourse(epoch_data, est, info, init, as_time=True, contours=0,vmin=-vm,vmax=vm, estimate_method=estimate_method)
        
        #print correlations
        if print_correlations and neve > 1:
            corr = np.corrcoef(topos[est_idx])[:-1,1:].diagonal()
            print(corr[~np.isnan(corr)])

def erp_data(epoched_data, times, channel,n_samples=None, pad=1):
    '''
    Create a data array compatible with the plot ERP function. Optionnally this function can resample the epochs to fit some provided times (e.g. onset of the events)

    Parameters
    ----------
     	epoched_data: xr.Dataset 
            Epoched physiological data with dims 'participant'X 'epochs' X 'channels'X 'samples'
        times: xr.Dataset
            Times between wich to extract or resample the data with dims 'trial_x_participant' X 'event'
        channel: str
            For which channel to extract the data
        n_samples: int
            How many samples to resample on if any
        pad: int
            padding added to the beginning and the end of the signal
    
    Returns
    -------
    data : nd.array
        array containing the extracted times for each epoch and stage with format epochs X events X samples.
    '''
    epoched_data = epoched_data.sel(channels=channel)
    if n_samples is None:
        data = np.zeros((len(times.trial_x_participant), len(times.event), len(epoched_data.samples)))*np.nan
    else:
        data = np.zeros((len(times.trial_x_participant), len(times.event), n_samples))*np.nan

    for i,trial in enumerate(times.trial_x_participant):
        for j,event in enumerate(times.event):
            if event == 0:
                sub_prevtime = 0
            else:
                sub_prevtime = times.sel(trial_x_participant=trial, event=event-1).data
            sub_times = times.sel(trial_x_participant=trial, event=event).data-1
            time_diff = sub_times-sub_prevtime
            if time_diff > 1:#rounds up to 0 otherwise
                subset = epoched_data.sel(trial_x_participant=trial, samples=slice(sub_prevtime,sub_times))
                if n_samples is None:
                    limit = np.sum(~subset.data.isnull()).values
                    data[i,j,:limit] = subset.data
                else:
                    padded_data = np.concatenate([np.repeat([subset.data[0]],pad), subset.data, np.repeat([subset.data[-1]], pad)])
                    #resample_poly seems to have better results than resample
                    data[i,j] = ssignal.resample_poly(padded_data,n_samples*10,len(padded_data.data)*(n_samples/10), padtype='median')
                    # data[i,j] = ssignal.resample(padded_data.data, n_samples)
    return data


def plot_erp(times, data, color='k',ax=None, minmax_lines=None, upsample=1, bootstrap=None, label=None):
    '''
    Plot the ERP based on the times extracted by HMP (or just stimulus and response and the data extracted from ```erp_data```.

    
    Parameters
    ----------
        times: xr.Dataset
            Times between wich to extract or resample the data with dims 
        data: nd.array
            numpy array from the erp_data functino
        color: str
            color for the lines
        ax: matplotlib.pyplot
            ax on which to draw
        minmax_lines: tuple
            Min and max arguments for the vertical lines on the plot
        upsample: float
            Upsampling factor for the times
        bootstrap: int
            how many bootstrap draw to perform
    '''
    if ax is None:
        ax = plt
    for event in times.event[1:]:
        time = times.sel(event=event-1).mean()
        time_current = int(times.sel(event=event).mean())
        if len(times.event)>2:
            x = np.linspace(time,time_current,num=np.shape(data)[-1])*upsample
            mean_signal =  np.nanmean(data[:,event,:],axis=0)
        else:
            x = np.arange(time,time_current)*upsample
            mean_signal =  np.nanmean(data[:,event,:],axis=0)[:time_current]
        ax.plot(x, mean_signal, color=color, label=label)
        if minmax_lines is not None:
            ax.vlines(times.mean('trial_x_participant')*upsample, minmax_lines[0],minmax_lines[1], color=color, ls=':', alpha=.25)
        if bootstrap is not None:
            test_boot = stats.bootstrap((data[:,event,:],), statistic=np.nanmean, n_resamples=bootstrap, axis=0, batch=10)
            if len(times.event)>2:
                ax.fill_between(x, test_boot.confidence_interval[0], test_boot.confidence_interval[1], alpha=0.5, color=color)
            else:
                ax.fill_between(x, test_boot.confidence_interval[0][:time_current], test_boot.confidence_interval[1][:time_current], alpha=0.5, color=color)



'''

'''

import os.path as op
import os
import numpy as np
import mne
from mne.datasets import sample
from warnings import warn
from copy import deepcopy

def available_sources(subselection=True):
    '''
    list available sources for sample subject in MNE
    '''
    data_path = sample.data_path()
    subjects_dir = op.join(data_path, 'subjects')
    labels = mne.read_labels_from_annot('sample', subjects_dir=subjects_dir, verbose=False)
    named_labels = []
    for label in range(len(labels)):
        named_labels.append(labels[label].name)
    named_labels = np.array(named_labels)
    return named_labels

def simulation_sfreq():
    data_path = sample.data_path()
    subjects_dir = op.join(data_path, 'subjects')
    evoked_fname = op.join(data_path, 'MEG', 'sample', 'sample_audvis-ave.fif')
    info = mne.io.read_info(evoked_fname, verbose=False)
    return info['sfreq']

def simulation_positions():
    data_path = sample.data_path()
    subjects_dir = op.join(data_path, 'subjects')
    subject = 'sample'
    evoked_fname = op.join(data_path, 'MEG', subject, 'sample_audvis-ave.fif')
    info = mne.io.read_info(evoked_fname, verbose=False)
    positions = np.delete(mne.channels.layout._find_topomap_coords(info, 'eeg'),52,axis=0)#inferring channel location using MNE    
    return positions

def simulation_info():
    data_path = sample.data_path()
    subjects_dir = op.join(data_path, 'subjects')
    subject = 'sample'
    evoked_fname = op.join(data_path, 'MEG', subject, 'sample_audvis-ave.fif')
    info = mne.io.read_info(evoked_fname, verbose=False)
    
    return info

def event_shape(event_width, event_width_samples, steps):
    '''
    Computes the template of a half-sine (event) with given frequency f and sampling frequency
    '''
    event_idx = np.arange(event_width_samples)*steps+steps/2
    event_frequency = 1000/(event_width*2)#gives event frequency given that events are defined as half-sines
    template = np.sin(2*np.pi*event_idx/1000*event_frequency)#event morph based on a half sine with given event width and sampling frequency
    template = template/np.sum(template**2)#Weight normalized
    return template

def simulate(sources, n_trials, n_jobs, file, relations=None, data_type='eeg', n_subj=1, path='./', overwrite=False, verbose=False, noise=True, times=None, seed=None, sfreq=100, save_snr=False, save_noiseless=False, event_length_samples=None, proportions=None):
    '''
    Simulates n_trials of EEG and/or MEG using MNE's tools based on the specified sources
    
    Parameters
    ----------
    sources : list
        2D or 3D list with dimensions (n_subjects *) sources * source_parameters
        Source parameters should contain :
        - the name of the source (see the output of available_source())
        - the duration of the event (in frequency, usually 10Hz)
        - the amplitude or strength of the signal from the source, expressed in volt (e.g. 1e-8 V)
        - the duration of the preceding stage as a scipy.stats distribution (e.g. scipy.stats.gamma(a, scale))
    n_trials: int
        Number of trials
    n_jobs: int
        Number of jobs to use with MNE's function (multithreading)
    file: str
        Name of the file to be saved (number of the subject will be added)
    relations: list
        list of int describing to which previous event is each event connected, 1 means stimulus, 2 means one event after stimulus, ... One event cannot be connected to an upcoming one
    data_type: str
        Whether to simulate "eeg" or "meg"
    n_subj: int
        How many subjects to simulate
    path: str
        path where to save the data
    overwrite: bool
        Whether to overwrite existing file
    verbose: bool
        Whether to display MNE's output
    noise: bool
        Adding noise to the simulated sources
    times: ndarray
        Deterministic simulation of event transitions times. Format is n_sources X n_trials
    location: float
        value in ms to add after a simulated event and before another one
    save_snr: bool
        Save the signal at peak value and electrode noise
    Returns
    -------
    generating_events: ndarray
        Times of the simulated events used to test for accurate recovery compared to estimation
    files: list
        list of file names (file + number of subject)
    '''
    if not verbose:
        mne.set_log_level('warning')
    else:
        mne.set_log_level(True)
    if seed is not None:
        random_state = np.random.RandomState(seed)
    else:
        random_state = np.random.RandomState(np.random.randint(low=0, high=3000))
    sources = np.array(sources, dtype=object)
    if len(np.shape(sources)) == 2:
        sources = [sources]#If only one subject
    if relations is None:
        relations = np.arange(len(sources[0])+1)+1
    if proportions is None:
        proportions = np.repeat(1, len(sources[0])+1)
    if np.shape(sources)[0] != n_subj:
        raise ValueError('Number of subject is not coherent with the sources provided')

    # Following code and comments largely comes from MNE examples (e.g. \
    # https://mne.tools/stable/auto_examples/simulation/simulated_raw_data_using_subject_anatomy.html)
    # It loads the data, info structure and forward solution for one example subject,
    # Note that all 'subject' will use this forward solution
    data_path = sample.data_path()
    subjects_dir = op.join(data_path, 'subjects')
    subject = 'sample'
    # First, we get an info structure from the test subject.
    evoked_fname = op.join(data_path, 'MEG', subject, 'sample_audvis-ave.fif')
    #mne read raw
    info = mne.io.read_info(evoked_fname, verbose=verbose)
    # To simulate sources, we also need a source space. It can be obtained from the
    # forward solution of the sample subject.
    fwd_fname = op.join(data_path, 'MEG', subject,'sample_audvis-meg-eeg-oct-6-fwd.fif')
    fwd = mne.read_forward_solution(fwd_fname, verbose=verbose)
    with info._unlock():
        info['sfreq'] = sfreq
    if data_type == 'eeg':
        picked_type = mne.pick_types(info, meg=False, eeg=True)
        fwd = mne.pick_types_forward(fwd, meg=False, eeg=True)
    elif data_type == 'meg':
        picked_type = mne.pick_types(info, meg=True, eeg=False)
        fwd = mne.pick_types_forward(fwd, meg=True, eeg=False)
    elif data_type == 'eeg/meg':
        picked_type = mne.pick_types(info, meg=True, eeg=True)
        fwd = mne.pick_types_forward(fwd, meg=True, eeg=True)
    else:
        raise ValueError(f'Invalid data type {data_type}, expected "eeg", "meg" or "eeg/meg"')
    src = fwd['src']
    info = mne.pick_info(info, picked_type)
    tstep = 1. / info['sfreq'] #sample duration

    
    #For each subject, simulate associated sources
    files = []
    for subj in range(n_subj):
        sources_subj = sources[subj]
        #Pre-allocate the random times to avoid generating too long 'recordings'
        
        rand_times = np.zeros((len(sources_subj),n_trials))
        random_indices_list = []
        for s, source in enumerate(sources_subj):
            #How many trials wil have the event, default is all
            props_trial = int(np.round(n_trials*proportions[s]))
            if proportions[s] < 1:
                random_indices_list.append(random_state.choice(np.arange(n_trials), size=props_trial, replace=False))
            else: random_indices_list.append(np.arange(n_trials))
            rand_times[s, random_indices_list[-1]] = source[-1].rvs(size=len(random_indices_list[-1]), random_state=random_state)
        seq_index = np.diff(relations, prepend=0)>0
        if seq_index.all():#If all events are sequential
            trial_time = np.sum(rand_times, axis=0)
        else:
            trial_time_seq =  np.sum(rand_times[seq_index], axis=0)
            trial_time_nonseq =  np.sum(rand_times[~seq_index], axis=0)
            trial_time = np.maximum(trial_time_seq, trial_time_nonseq)
        trial_time /= 1000 #to seconds
        trial_time[1:] += trial_time[:-1]+.5
        trial_time = np.cumsum(trial_time)
        #Build simulator
        files_subj = []
        source_simulator = mne.simulation.SourceSimulator(src, tstep=tstep, first_samp=0, \
                    duration=trial_time[-1]+10)#add 10sec to the end of the last trial
        if n_subj == 1: subj_file = file + f'_raw.fif'
        else: subj_file = file + f'_{subj}_raw.fif'
        if subj_file in os.listdir(path) and not overwrite:
            subj_file = path+subj_file
            warn(f'{subj_file} exists no new simulation performed', UserWarning)
            files_subj.append(subj_file)
            files_subj.append(subj_file.split('.fif')[0]+'_generating_events.npy')
            if save_snr:
                files_subj.append(subj_file.split('.fif')[0]+'_snr.npy')
            if save_noiseless:
                files_subj.append(file + f'_noiseless_raw.fif')
            files.append(files_subj)
        else:
            subj_file = op.join(path, subj_file)
            print(f'Simulating {subj_file}')
            # stim_onset occurs every x samples.
            events = np.zeros((n_trials, 3), int)
            stim_onsets =  1 + trial_time #offset of first stim is 1 s
            events[:,0] = stim_onsets/tstep#last event 
            events[:,2] = 1#trigger 1 = stimulus 

            #Fake source, actually stimulus onset
            selected_label = mne.read_labels_from_annot(
                    subject, regexp=sources_subj[0][0], subjects_dir=subjects_dir, verbose=verbose)[0]
            label = mne.label.select_sources(subject, selected_label, subjects_dir=subjects_dir, random_state=random_state)
            source_time_series = np.array([1e-20])#stim trigger
            source_simulator.add_data(label, source_time_series, events)

            trigger = 2
            generating_events = events
            for s, source in enumerate(sources_subj):
                if trigger == len(sources_subj)+1:
                    source[2] = 1e-20#Last source defines RT and is not an event per se
                selected_label = mne.read_labels_from_annot(
                    subject, regexp=source[0], subjects_dir=subjects_dir, verbose=verbose)[0]
                label = mne.label.select_sources(subject, selected_label, subjects_dir=subjects_dir, location=0, grow_outside=False, random_state=random_state)
                #last two parameters ensure sources that are different enough
                # Define the time course of the activity for each source of the region to
                # activate
                if event_length_samples is None:
                    event_duration = int(((1/source[1])/2)*info['sfreq'])
                    #Shift the trigger for the simulations 
                    shift = event_duration//2
                else:
                    event_duration = event_length_samples[s]
                    # Assumes the shortest event duration is a half-sine
                    shift = min(event_length_samples)//2
                source_time_series = event_shape(((1000/source[1])/2),event_duration,1000/info['sfreq']) * source[2]
                
                #adding source event, take as previous time the event defined by relation (default is previous event)
                events = generating_events[generating_events[:,-1] == relations[s]].copy()
                random_indices = random_indices_list[s]
                if times is None:
                    # Always at least one sample later
                    rand_i = np.maximum(1,rand_times[s]/(tstep*1000))
                    rand_i = np.round(rand_i,decimals=0)
                else:
                    rand_i = times[:,trigger-2]/(tstep*1000)
                if len(rand_i[rand_i<0]) > 0:
                    warn(f'Negative stage duration were found, 1 is imputed for the {len(rand_i[rand_i<0])} trial(s)', UserWarning)
                    rand_i[rand_i<0] = 1
                events[random_indices, 0] = events[random_indices,0] + rand_i[random_indices] # Events sample.
                events[:, 2] = trigger  # All events have the sample id.
                trigger += 1
                generating_events = np.concatenate([generating_events, events.copy()])
                #Shift event to onset when simulating pattern
                events[random_indices, 0] = events[random_indices,0] - shift
                #add these events
                source_simulator.add_data(label, source_time_series, events[random_indices])

            generating_events = generating_events[generating_events[:, 0].argsort()]
            # Project the source time series to sensor space and add some noise. The source
            # simulator can be given directly to the simulate_raw function.
            raw = mne.simulation.simulate_raw(info, source_simulator, forward=fwd, n_jobs=n_jobs,verbose=verbose)
            
            n_events = len(sources_subj)-1
            if save_noiseless:
                raw.save(file + f'_noiseless_raw.fif', overwrite=True)
            if data_type == 'eeg':
                raw = raw.pick_types(meg=False, eeg=True, stim=True)
            elif data_type == 'meg':
                raw = raw.pick_types(meg=True, eeg=False, stim=True)
            elif data_type == 'eeg/meg':
                raw = raw.pick_types(meg=True, eeg=True, stim=True)
            if save_snr:
                snr = np.zeros((len(info['ch_names']), n_events))
                data = deepcopy(raw.get_data())
                for event in range(n_events):
                    times_out = generating_events[generating_events[:,2] == event+2,0]
                    snr[:,event] = np.mean((data[:, times_out])**2, axis=-1)
            if noise:
                cov = mne.make_ad_hoc_cov(raw.info, verbose=verbose)
                mne.simulation.add_noise(raw, cov,  verbose=verbose,iir_filter=[0.2, -0.2, 0.04], random_state=random_state)

            raw.save(subj_file, overwrite=True)
            files_subj.append(subj_file)
            np.save(subj_file.split('.fif')[0]+'_generating_events.npy', generating_events)
            files_subj.append(subj_file.split('.fif')[0]+'_generating_events.npy')
            if save_snr:
                data = raw.get_data()
                for event in range(n_events):
                    times_out = generating_events[generating_events[:,2] == event+2,0]
                    snr[:,event] /= np.var(data[:, times_out], axis=-1)
                np.save(subj_file.split('.fif')[0]+'_snr.npy', snr)
                files_subj.append(subj_file.split('.fif')[0]+'_snr.npy')
            files.append(files_subj)
            if save_noiseless:
                files_subj.append(file + f'_noiseless_raw.fif')
            print(f'{subj_file} simulated')
            
    if n_subj == 1:
        files = files[0]
    return files
        

def demo(cpus, n_events, seed=123):
    
    ## Imports and code specific to the simulation (see tutorial 3 and 4 for real data)
    from scipy.stats import gamma
    import matplotlib.pyplot as plt 
    from hmp.utils import read_mne_data
    
    random_gen =  np.random.default_rng(seed=seed)

    ## Parameters for the simulations
    frequency, amplitude = 10., .1e-6 #Frequency of the transition event and its amplitude in Volt
    shape = 2#shape of the gamma distribution

    #Storing electrode position, specific to the simulations
    positions = simulation_info()#Electrode position
    sfreq = 100
    all_source_names = available_sources()#all brain sources you can play with
    n_trials = 50 #Number of trials to simulate
    
    # Randomly specify the transition events
    name_sources = random_gen.choice(all_source_names,n_events+1, replace=False)#randomly pick source without replacement
    times = np.array([100,150,200,50,50,50,150,200,50])/shape #designed to fail with default starting points

    sources = []
    for source in range(len(name_sources)):
        sources.append([name_sources[source], frequency, amplitude, \
                          gamma(shape, scale=times[source])]) #gamma returns times in ms

    file = 'dataset_tutorial2' #Name of the file to save

    #Simulating and recover information on electrode location and true time onset of the simulated events
    files = simulate(sources, n_trials, cpus,file, overwrite=False, seed=seed, noise=True, sfreq=sfreq)
    
    generating_events = np.load(files[1])
    #events_resamp = generating_events.copy()
    #events_resamp[:, 0] = events_resamp[:, 0] * float(resample_freq) / sfreq

    number_of_sources = len(np.unique(generating_events[:,2])[1:])#one trigger = one source
    random_source_times = np.reshape(np.diff(generating_events[:,0], prepend=0),(n_trials,number_of_sources+1))[:,1:] #By-trial generated event times

    #Reading the necessary info to read the EEG data
    resp_trigger = int(np.max(np.unique(generating_events[:,2])))#Resp trigger is the last source in each trial
    event_id = {'stimulus':1}
    resp_id = {'response':resp_trigger}
    events = generating_events[(generating_events[:,2] == 1) | (generating_events[:,2] == resp_trigger)]#only retain stimulus and response triggers

    # Reading the data
    eeg_dat = read_mne_data(files[0], event_id, resp_id, events_provided=events, verbose=False)
    
    all_other_chans = range(len(positions.ch_names[:-61]))#non-eeg
    chan_list = list(np.arange(len(positions.ch_names)))
    chan_list = [e for e in chan_list if e not in all_other_chans]
    chan_list.pop(52)#Bad elec
    positions = mne.pick_info(positions, sel=chan_list)
    return eeg_dat, random_source_times, positions

def classification_true(true_topologies,test_topologies):
    '''
    Classifies event as belonging to one of the true events

    Parameters,
    ----------
    true_topologies : xarray.DataArray
        topologies for the true events simulated obtained from `init.compute_topologies(epoch_data, test_estimates, test_init, mean=True)`
    test_tolopogies : xarray.DataArray
        topologies for the events found in the estimation procedure obtained from `init.compute_topologies(epoch_data, true_estimates, true_init, mean=True)`

    
    Returns
    -------
    idx_true_positive: np.array
        index of the true events found in the test estimation
    corresp_true_idx: np.array
        index in the test estimate that correspond to the indexes in corresp_true_idx
        
    '''
    test_topologies = (test_topologies.copy()-test_topologies.mean(axis=1))/ test_topologies.std(axis=1)
    true_topologies = (true_topologies.copy()-true_topologies.mean(axis=1))/true_topologies.std(axis=1)
    true0 = np.vstack((np.zeros(true_topologies.shape[1]), true_topologies))#add a zero electrode event
    classif = np.zeros(test_topologies.shape[0], dtype=int)#array of categorization in true events
    classif_vals = np.zeros(test_topologies.shape[0])#values of the squared diff
    for i, test_ev in enumerate(test_topologies):
        all_distances = np.zeros(len(true0))
        for j, true_ev in enumerate(true0):
            all_distances[j] = np.median(np.abs(true_ev-test_ev))
        classif[i] = np.argmin(all_distances)
        classif_vals[i] = all_distances[classif[i]]
        
    mapping_true = {}
    for test_idx, (idx, val) in enumerate(zip(classif, classif_vals)):
        if idx > 0:
            if idx not in mapping_true or val < mapping_true[idx]:
                mapping_true[idx] = test_idx
                
    corresp_true_idx = np.array(list(mapping_true.keys()))-1#Corresponding true index, excluding 0 event
    idx_true_positive = np.array(list(mapping_true.values()))
    return idx_true_positive, corresp_true_idx 
    
def simulated_times_and_parameters(generating_events, init, resampling_freq=None, data=None):
    sfreq = init.sfreq
    n_stages = len(np.unique(generating_events[:,2])[1:])#one trigger = one source
    n_events = n_stages-1 
    if resampling_freq is None:
        resampling_freq = sfreq

    #Recover the actual time of the simulated events
    random_source_times = np.zeros((int(len(generating_events)/(n_stages+1)), n_stages))
    i,x = 1,0                  
    while x < len(random_source_times):
        for j in np.arange(n_stages):#recovering the individual duration- of event onset
            random_source_times[x,j] = generating_events[i,0] - generating_events[i-1,0]
            i += 1
        i += 1
        x += 1
        
    ## Recover parameters
    true_parameters = np.tile(init.shape, (n_stages, 2))
    true_parameters[:,1] = init.mean_to_scale(np.mean(random_source_times,axis=0),init.shape)
    true_parameters[true_parameters[:,1] <= 0, 1] = 1e-3#Can happen in corner cases
    random_source_times = random_source_times*(1000/sfreq)/(1000/resampling_freq)
    ## Recover magnitudes
    sample_times = np.zeros((init.n_trials, n_events), dtype=int)
    for event in range(n_events):
        for trial in range(init.n_trials):
            trial_time = init.starts[trial]+np.sum(random_source_times[trial,:event+1])
            if init.ends[trial] >= trial_time:#exceeds RT
                sample_times[trial,event] = trial_time
            else:
                sample_times[trial,event] = init.ends[trial]
    if data is None: #use crosscorrelated data
        true_activities = init.events[sample_times[:,:]]
    else:
        true_activities = data[sample_times[:,:]]
    true_magnitudes = np.mean(true_activities, axis=0)
    return random_source_times.astype(int), true_parameters, true_magnitudes, true_activities


""" 
Adapted from https://github.com/ANCPLabOldenburg/MCCA/blob/main/README.md
Author/maintainer: Leo Michalke https://github.com/lmichalke


MIT License

Copyright (c) 2022 Applied Neurocognitive Psychology Lab University Oldenburg

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import numpy as np
from sklearn.decomposition import PCA
from sklearn.exceptions import NotFittedError
from scipy.linalg import eigh, norm
import warnings



class MCCA:
    """ Performs multiset canonical correlation analysis with an optional
        regularization term based on spatial similarity of weight maps. The
        stronger the regularization, the more similar weight maps are forced to
        be across subjects. Note that the term 'weights' is used interchangeably
        with PCA / MCCA eigenvectors here.

    Parameters:
        n_components_pca (int): Number of PCA components to retain for each subject (default 50)

        n_components_mcca (int): Number of MCCA components to retain (default 10)

        cov (bool): If true (default), apply the PCA to the var-cov matrix
        
        r (int or float): Regularization strength (default 0)

        pca_only (bool): If true, skip MCCA calculation (default False)

    Attributes:
        mu (ndarray): Mean subtracted before PCA (subjects, sensors)

        sigma (ndarray): PCA standard deviation (subjects, PCs)

        pca_weights (ndarray): PCA weights that transform sensors to PCAs for each
                               subject (subjects, sensors, PCs)

        mcca_weights (ndarray): MCCA weights that transform PCAs to MCCAs for each subject.
                                None if pca_only is True. (subjects, PCs, CCs)
    """

    def __init__(self, n_components_pca, n_components_mcca, r=0, pca_only=False):
        if n_components_mcca > n_components_pca:
            import warnings
            warnings.warn(f"Warning........... number of MCCA components ({n_components_mcca}) cannot be greater than "
                          f"number of PCA components ({n_components_pca}), setting them equal.")
            n_components_mcca = n_components_pca
        self.n_pcs = n_components_pca
        self.n_ccs = n_components_mcca
        self.r = r
        self.pca_only = pca_only
        self.mcca_weights, self.pca_weights, self.mu, self.sigma = None, None, None, None

    def obtain_mcca(self, X):
        """ Apply individual-subject PCA and across-subjects MCCA.

        Parameters:
            X (ndarray): Input data in sensor space (subjects, samples, sensors)

        Returns:
            scores (ndarray): Returns scores in PCA space if self.pca_only is true and MCCA scores otherwise.
        """
        n_subjects, n_samples, n_sensors = X.shape
        X_pca = np.zeros((n_subjects, n_samples, self.n_pcs))
        self.pca_weights = np.zeros((n_subjects, n_sensors, self.n_pcs))
        self.mu = np.zeros((n_subjects, n_sensors))
        self.sigma = np.zeros((n_subjects, self.n_pcs))
        lim = 0
        # obtain subject-specific PCAs
        for i in range(n_subjects):
            pca = PCA(n_components=self.n_pcs, svd_solver='full')
            x_i = np.squeeze(X[i]).copy()  # time x sensors
            score = pca.fit_transform( x_i[~np.isnan(x_i[:,0]),:])
            self.mu[i] = pca.mean_
            self.sigma[i] = np.sqrt(pca.explained_variance_)
            score /= self.sigma[i]
            lim_i = len(x_i[~np.isnan(x_i[:,0])])
            lim = int(np.max([lim, lim_i]))
            self.pca_weights[i] = pca.components_.T
            X_pca[i,:lim_i,:] = score
        warnings.warn(f'MCCA is done on {lim} samples per subject')
        X_pca = X_pca[:, :lim,:]
        
        if self.pca_only:
            return X_pca
        else:
            return self._mcca(X_pca)

    def obtain_mcca_cov(self, X):
        """ Apply individual-subject PCA on the variance-covariance matrix and across-subjects MCCA

        Parameters:
            X (ndarray): Input data in sensor space for each trial (subjects, n_trials, samples, sensors)

        Returns:
            scores (ndarray): Returns scores in PCA space if self.pca_only is true and MCCA scores otherwise.
        """
        n_subjects, n_trials, n_samples, n_sensors = X.shape
        X_pca = np.zeros((n_subjects, n_trials*n_samples, self.n_pcs))
        self.pca_weights = np.zeros((n_subjects, n_sensors, self.n_pcs))
        lim = 0
        self.mu = np.zeros((n_subjects, n_sensors))
        self.sigma = np.ones((n_subjects, self.n_pcs))
        # obtain subject-specific PCAs
        for i in range(n_subjects):
            pca = PCA(n_components=self.n_pcs, svd_solver='full')
            x_i = np.squeeze(X[i]).copy()  # time x sensors
            av_vcov = np.mean(\
                [np.cov(
                    x_i[trial,~np.isnan(x_i[trial,:,0]),:]
                    .T)\
                for trial in range(x_i.shape[0]) if ~np.isnan(x_i[trial,:,0]).all()],axis=0)
            pca.fit(av_vcov)
            x_i = x_i.reshape(n_trials*n_samples, n_sensors)
            score =  x_i[~np.isnan(x_i[:,0]),:] @ pca.components_.T
            lim_i = len(x_i[~np.isnan(x_i[:,0])])
            lim = int(np.max([lim, lim_i]))
            self.pca_weights[i] = pca.components_.T
            X_pca[i,:lim_i,:] = score
        warnings.warn(f'MCCA is done on {lim} out of {n_trials*n_samples} samples per subject')
        X_pca = X_pca[:, :lim,:]
        
        if self.pca_only:
            return X_pca
        else:
            return self._mcca(X_pca)

    
    def _mcca(self, pca_scores):
        """ Performs multiset canonical correlation analysis with an optional
            regularization term based on spatial similarity of weight maps. The
            stronger the regularization, the more similar weight maps are forced to
            be across subjects.

        Parameters:
            pca_scores (ndarray): Input data in PCA space (subjects, samples, PCs)

        Returns:
            mcca_scores (ndarray): Input data in MCCA space (subjects, samples, CCs).
        """
        # R_kl is a block matrix containing all cross-covariances R_kl = X_k^T X_l between subjects k, l, k != l
        # where X is the data in the subject-specific PCA space (PCA scores)
        # R_kk is a block diagonal matrix containing auto-correlations R_kk = X_k^T X_k in its diagonal blocks
        R_kl, R_kk = _compute_cross_covariance(pca_scores)
        # Regularization
        if self.r != 0:
            # The regularization terms W_kl and W_kk are calculated the same way as R_kl and R_kk above, but using
            # cross-covariance of PCA weights instead of PCA scores
            W_kl, W_kk = _compute_cross_covariance(self.pca_weights)
            # Add regularization term to R_kl and R_kk
            R_kl += self.r * W_kl
            R_kk += self.r * W_kk
        # Obtain MCCA solution by solving the generalized eigenvalue problem
        #                   R_kl h = p R_kk h
        # where h are the concatenated eigenvectors of all subjects and
        # p are the generalized eigenvalues (canonical correlations).
        # If PCA scores are whitened and no regularisation is used, R_kk is an identity matrix and the generalized
        # eigenvalue problem is reduced to a regular eigenvalue problem
        p, h = eigh(R_kl, R_kk, subset_by_index=(R_kl.shape[0] - self.n_ccs, R_kl.shape[0] - 1))
        # eigh returns eigenvalues in ascending order. To pick the k largest from a total of n eigenvalues,
        # we use subset_by_index=(n - k, n - 1).
        # Flip eigenvectors so that they are in descending order
        h = np.flip(h, axis=1)
        # Reshape h from (subjects * PCs, CCs) to (subjects, PCs, CCs)
        h = h.reshape((pca_scores.shape[0], self.n_pcs, self.n_ccs))
        # Normalize eigenvectors per subject
        self.mcca_weights = h / norm(h, ord=2, axis=(1, 2), keepdims=True)
        return np.matmul(pca_scores, self.mcca_weights)

    def transform_trials(self, X, subject=0):
        """ Use of MCCA weights (obtained from averaged data) to transform single
            trial data from sensor space to MCCA space.

        Parameters:
            X (ndarray): Single trial data of one subject in sensor space
                         (trials, samples, sensors)
            subject (int): Index of the subject whose data is being transformed

        Returns:
            X_mcca (ndarray): Transformed single trial data in MCCA space
                            (trials, samples, CCs)
        """
        if self.mcca_weights is None:
            raise NotFittedError('MCCA needs to be fitted before calling transform_trials')
        X -= self.mu[np.newaxis, np.newaxis, subject]  # centered
        X_pca = np.matmul(X, self.pca_weights[subject])
        X_pca /= self.sigma[np.newaxis, np.newaxis, subject]  # standardized
        X_mcca = np.matmul(X_pca, self.mcca_weights[subject])
        return X_mcca

    def inverse_transform_trials(self, X_mcca, subject=0):
        """ Transform single trial data from MCCA space back to sensor space.

        Parameters:
            X_mcca (ndarray): Single trial data of one subject in MCCA space
                            (trials, samples, CCs)
            subject (int): Index of the subject whose data is being transformed

        Returns:
            X (ndarray): Single trial data transformed back into sensor space
                         (trials, samples, sensors)
        """
        if self.mcca_weights is None:
            raise NotFittedError('MCCA needs to be fitted before calling inverse_transform_trials')
        X_pca = np.matmul(X_mcca, np.linalg.pinv(self.mcca_weights[subject]))  # trials x samples x PCAs
        X_pca *= self.sigma[np.newaxis, np.newaxis, subject]  # revert standardization
        X = np.matmul(X_pca, np.linalg.pinv(self.pca_weights[subject]))
        X += self.mu[np.newaxis, np.newaxis, subject]  # revert centering
        return X

    def transform_trials_pca(self, X, subject=0):
        """ Transform single trial data from sensor space to PCA space

        Parameters:
            X (ndarray): Single trial data of one subject in sensor space
                         (trials, samples, sensors)
            subject (int): Index of the subject whose data is being transformed

        Returns:
            X_pca (ndarray): Transformed single trial data in PCA space
                             (trials, samples, PCs)
        """
        if self.pca_weights is None:
            raise NotFittedError('PCA needs to be fitted before calling transform_trials_pca')
        X -= self.mu[np.newaxis, np.newaxis, subject]  # centered
        X_pca = np.matmul(X, self.pca_weights[subject])
        X_pca /= self.sigma[np.newaxis, np.newaxis, subject]  # standardized
        return X_pca

    def test_mcca(self, data_train, data_test):
        """ Test if the inter-subject correlations/consistency observed from training
            data MCCAs generalize over testing data.

        Parameters:
            data_train (ndarray): Training data in sensor space
                                  (subjects, samples, sensors)
            data_test (ndarray): Test data in sensor space
                                 (subjects, samples, sensors)

        Returns:
            correlations (ndarray): Inter-subject correlations (averaged over
                                    every pair of subjects) over every MCCA component
                                    (first row: training data; second row: testing data)
        """
        if self.mcca_weights is None:
            raise NotFittedError('MCCA needs to be fitted before calling test_mcca')
        t_mu = self.mu[:, np.newaxis]
        t_sigma = self.sigma[:, np.newaxis]
        data_train -= t_mu
        pca_train = np.matmul(data_train, self.pca_weights)
        pca_train /= t_sigma
        mcca_train = np.matmul(pca_train, self.mcca_weights)
        data_test -= t_mu
        pca_test = np.matmul(data_test, self.pca_weights)
        pca_test /= t_sigma
        mcca_test = np.matmul(pca_test, self.mcca_weights)

        K = self.mcca_weights.shape[2]
        correlations = np.zeros((K, 2))
        for i in range(K):
            inter_subject_corr_train = np.corrcoef(np.squeeze(mcca_train[:, :, i]))
            inter_subject_corr_test = np.corrcoef(np.squeeze(mcca_test[:, :, i]))
            # averaged inter-subject correlations over training data
            correlations[i, 0] = np.mean(inter_subject_corr_train)
            # averaged inter-subject correlations over testing data
            correlations[i, 1] = np.mean(inter_subject_corr_test)

        return correlations


def _compute_cross_covariance(X):
    """ Computes cross-covariance of PCA scores or components between subjects.

    Parameters:
        X (ndarray): PCA scores (subjects, samples, PCs) or weights (subjects, sensors, PCs)

    Returns:
        R_kl (ndarray): Block matrix containing all cross-covariances R_kl = X_k^T X_l between subjects k, l, k != l
                        with shape (subjects * PCs, subjects * PCs)
        R_kk (ndarray): Block diagonal matrix containing auto-correlations R_kk = X_k^T X_k in its diagonal blocks
                        with shape (subjects * PCs, subjects * PCs)
    """
    n_subjects, n_samples, n_pcs = X.shape
    R = np.cov(X.swapaxes(1, 2).reshape(n_subjects * n_pcs, n_samples))
    R_kk = R * np.kron(np.eye(n_subjects), np.ones((n_pcs, n_pcs)))
    R_kl = R - R_kk
    return R_kl, R_kk
